
from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Sequence, Iterable, IO, Mapping, Union
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.submission_ASYNC import Submission, TextPost
    from ...models.comment_ASYNC import Comment
    from ...dtos.submission import GalleryItem
    from ...types import JSON_ro

import os.path as op
from functools import cached_property
import json

from ...model_loaders.submission_ASYNC import load_submission, load_text_post
from ...models.media_upload_lease import MediaUploadLease
from ...model_loaders.media_upload_lease import load_media_upload_lease
from ...http.util.guess_filename_mimetype import guess_filename_mimetype
from ...util.base_conversion import to_base36
from ...iterators.chunking import chunked
from ...iterators.call_chunk_calling_async_iterator import CallChunkCallingAsyncIterator
from ...iterators.call_chunk_chaining_async_iterator import CallChunkChainingAsyncIterator
from ...iterators.async_call_chunk import AsyncCallChunk
from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.submission_async1 import SubmissionSearchAsyncPaginator, SubmissionDuplicatesAsyncPaginator
from ...model_loaders.comment_ASYNC import load_comment
from .fetch_ASYNC import Fetch
from .get_ASYNC import Get

class SubmissionProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client
        self.fetch: Fetch = Fetch(self, client)
        ("""
            Fetch a submission.

            .. .PARAMETERS

            :param `int` idn:
                Submission ID.

            .. .RETURNS

            :rtype: :class:`~.models.submission_ASYNC.Submission`

            .. .RAISES

            :raises redditwarp.exceptions.NoResultException:
                The target was not found.
            """)
        self.get: Get = Get(client)
        ("""
            Get a submission.

            .. .PARAMETERS

            :param `int` idn:
                Submission ID.

            .. .RETURNS

            :rtype: `Optional`\\[:class:`~.models.submission_ASYNC.Submission`]
            """)

    def bulk_fetch(self, ids: Iterable[int]) -> CallChunkChainingAsyncIterator[Submission]:
        """Bulk fetch submissions.

        Any ID not found will be ignored.

        .. .PARAMETERS

        :param `Iterable[int]` ids:
            Submission IDs.

        .. .RETURNS

        :rtype: :class:`~.iterators.call_chunk_chaining_async_iterator.CallChunkChainingAsyncIterator`\\[:class:`~.models.submission_ASYNC.Submission`]
        """
        async def mass_fetch(ids: Sequence[int]) -> Sequence[Submission]:
            id36s = map(to_base36, ids)
            full_id36s = map('t3_'.__add__, id36s)
            ids_str = ','.join(full_id36s)
            root = await self._client.request('GET', '/api/info', params={'id': ids_str})
            return [load_submission(i['data'], self._client) for i in root['data']['children']]

        return CallChunkChainingAsyncIterator(AsyncCallChunk(mass_fetch, chunk) for chunk in chunked(ids, 100))

    async def reply(self, idn: int, body: Union[str, Mapping[str, JSON_ro]]) -> Comment:
        """Comment on a submission.

        .. .PARAMETERS

        :param `int` idn:
        :param body:
            Either markdown or richtext.
        :type body: `Union`\\[`str`, `Mapping`\\[`str`, :class:`~.types.JSON_ro`]]

        .. .RETURNS

        :rtype: :class:`~.models.comment_ASYNC.Comment`

        .. .RAISES

        :(raises):
            Same as comment :meth:`~.siteprocs.comment.SYNC.CommentProcedures.reply`
            but for submissions.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('thing_id', 't3_' + to_base36(idn))
            yield ('return_rtjson', '1')
            if isinstance(body, str):
                yield ('text', body)
            else:
                yield ('richtext_json', json.dumps(body))

        result = await self._client.request('POST', '/api/comment', files=dict(g()))
        return load_comment(result, self._client)

    class MediaUploading:
        def __init__(self, outer: SubmissionProcedures) -> None:
            self._client = outer._client

        async def __call__(self,
            file: IO[bytes],
            *,
            filepath: Optional[str] = None,
            timeout: float = 1000,
        ) -> MediaUploadLease:
            return await self.upload(file, filepath=filepath, timeout=timeout)

        async def obtain_upload_lease(self,
            *,
            filepath: str,
            mimetype: Optional[str] = None,
        ) -> MediaUploadLease:
            if mimetype is None:
                mimetype = guess_filename_mimetype(filepath)
            result = await self._client.request('POST', '/api/media/asset',
                    data={'filepath': filepath, 'mimetype': mimetype})
            return load_media_upload_lease(result)

        async def deposit_file(self,
            file: IO[bytes],
            upload_lease: MediaUploadLease,
            *,
            timeout: float = 1000,
        ) -> None:
            resp = await self._client.http.request('POST', upload_lease.endpoint,
                    data=upload_lease.fields, files={'file': file}, timeout=timeout)
            resp.ensure_successful_status()

        async def upload(self,
            file: IO[bytes],
            *,
            filepath: Optional[str] = None,
            timeout: float = 1000,
        ) -> MediaUploadLease:
            if filepath is None:
                filepath = op.basename(getattr(file, 'name', ''))
                if not filepath:
                    raise ValueError("the `filepath` parameter must be explicitly specified if the file object has no `name` attribute.")
            upload_lease = await self.obtain_upload_lease(filepath=filepath)
            await self.deposit_file(file, upload_lease, timeout=timeout)
            return upload_lease

    media_uploading: cached_property[MediaUploading] = cached_property(MediaUploading)

    async def create_text_post(self,
        sr: str,
        title: str,
        body: Union[str, Mapping[str, JSON_ro]],
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        oc: bool = False,
        collection_uuid: Optional[str] = None,
        flair_uuid: Optional[str] = None,
        flair_text: Optional[str] = None,
        event_start: Optional[str] = None,
        event_end: Optional[str] = None,
        event_tz: Optional[str] = None,
    ) -> int:
        """Create a text post.

        .. .PARAMETERS

        :param `str` sr:
            Name of the subreddit to submit the post to.
        :param `str` title:
            Title of the post.
        :param body:
            The body text of the post.

            Specify either markdown text or a richtext document.
        :type body: `Union`\\[`str`, `Mapping`\\[`str`, :class:`~.types.JSON_ro`]]
        :param `bool` reply_notifications:
            Receive inbox notifications for replies.
        :param `bool` spoiler:
            Mark as spoiler.
        :param `bool` nsfw:
            Mark as NSFW.
        :param `bool` oc:
            Mark as original content.
        :param `Optional[str]` collection_uuid:
            The UUID of a collection to add this post to a collection.
        :param `Optional[str]` flair_uuid:
            The UUID of a flair template to use.
        :param `Optional[str]` flair_text:
            Custom flair text.
        :param `Optional[str]` event_start:
            A datetime ISO 8601 string. E.g. `2018-09-11T12:00:00`.
        :param `Optional[str]` event_end:
            A datetime ISO 8601 string.
        :param `Optional[str]` event_tz:
            A timezone. E.g., `America/Los_Angeles`.

        .. .RETURNS

        :returns: The ID of the newly created post.
        :rtype: `int`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `BAD_SR_NAME`:
                An empty string was specified for `sr`.
            + `SUBREDDIT_NOEXIST`:
               - The specified subreddit does not exist.
               - The specified subreddit is invalid.

            + `SUBREDDIT_NOTALLOWED`:
               - The subreddit is restricted and you are not an approved user.
               - You are banned from the subreddit.
               - You are trying to submit an image or video post to a NSFW subreddit.

               Note, quarantined subreddits can be posted to even if you haven't
               yet opt-ed in to viewing its content.
            + `NO_TEXT`:
                The `title` parameter was not specified, was blank, or contained only whitespace.
            + `JSON_PARSE_ERROR`:
                Richtext was passed and it was not in the correct format.
            + `TOO_LONG`:
               - The `title` parameter must be under 300 characters.
               - The `body` parameter must be under 40000 characters.

            + `NO_SELFS`:
                The subreddit doesn't accept text posts.
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The target subreddit is private or banned.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('kind', 'self')
            yield ('sr', sr)
            yield ('title', title)
            if isinstance(body, str):
                yield ('text', body)
            else:
                yield ('richtext_json', json.dumps(body))
            yield ('sendreplies', '01'[reply_notifications])
            if spoiler: yield ('spoiler', '1')
            if nsfw: yield ('nsfw', '1')
            if oc: yield ('original_content', '1')
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        root = await self._client.request('POST', '/api/submit', files=dict(g()))
        return int(root['json']['data']['id'], 36)

    async def create_link_post(self,
        sr: str,
        title: str,
        link: str,
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        oc: bool = False,
        collection_uuid: Optional[str] = None,
        flair_uuid: Optional[str] = None,
        flair_text: Optional[str] = None,
        event_start: Optional[str] = None,
        event_end: Optional[str] = None,
        event_tz: Optional[str] = None,
        resubmit: bool = True,
    ) -> int:
        """Create a link post.

        Behaves similarly to :meth:`.create_text_post`.

        .. .PARAMETERS

        :(parameters): Similar to :meth:`.create_text_post`.

        :param `str` url:
            A URL.
        :param `bool` resubmit:
            When the "Restrict how often the same link can be posted" content control
            setting is enabled, if a link with the same URL has already been submitted
            then an `ALREADY_SUB` API error would be returned unless this field is true.

        .. .RETURNS

        :(returns): Similar to :meth:`.create_text_post`.

        .. .RAISES

        :(raises): Similar to :meth:`.create_text_post`.

        :raises redditwarp.exceptions.RedditError:
            + `NO_URL`:
                The `link` parameter was not specified, or the URL is invalid.
            + `ALREADY_SUB`:
                The given link has already been submitted to the subreddit.
                See parameter `resubmit`.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('kind', 'link')
            yield ('sr', sr)
            yield ('title', title)
            yield ('url', link)
            if resubmit: yield ('resubmit', '1')
            yield ('sendreplies', '01'[reply_notifications])
            if spoiler: yield ('spoiler', '1')
            if nsfw: yield ('nsfw', '1')
            if oc: yield ('original_content', '1')
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        root = await self._client.request('POST', '/api/submit', data=dict(g()))
        return int(root['json']['data']['id'], 36)

    async def create_image_post(self,
        sr: str,
        title: str,
        link: str,
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        oc: bool = False,
        collection_uuid: Optional[str] = None,
        flair_uuid: Optional[str] = None,
        flair_text: Optional[str] = None,
        event_start: Optional[str] = None,
        event_end: Optional[str] = None,
        event_tz: Optional[str] = None,
    ) -> None:
        """Create an image post.

        Behaves similarly to :meth:`.create_text_post`.

        .. .PARAMETERS

        :(parameters): Similar to :meth:`.create_text_post`.

        :param `str` link:
            A URL to an image.

        .. .RETURNS

        :(returns): Similar to :meth:`.create_text_post`.

        .. .RAISES

        :(raises): Similar to :meth:`.create_text_post`.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('kind', 'image')
            yield ('sr', sr)
            yield ('title', title)
            yield ('url', link)
            yield ('sendreplies', '01'[reply_notifications])
            if spoiler: yield ('spoiler', '1')
            if nsfw: yield ('nsfw', '1')
            if oc: yield ('original_content', '1')
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        await self._client.request('POST', '/api/submit', data=dict(g()))

    async def create_video_post(self,
        sr: str,
        title: str,
        link: str,
        thumbnail: str,
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        oc: bool = False,
        collection_uuid: Optional[str] = None,
        flair_uuid: Optional[str] = None,
        flair_text: Optional[str] = None,
        event_start: Optional[str] = None,
        event_end: Optional[str] = None,
        event_tz: Optional[str] = None,
        vgif: bool = False,
    ) -> None:
        """Create a video post.

        Behaves similarly to :meth:`.create_text_post`.

        .. .PARAMETERS

        :(parameters): Similar to :meth:`.create_text_post`.

        :param `str` link:
            A URL to a video.
        :param `str` thumbnail:
            A URL to an image to be used as a thumbnail for the video.
        :param `bool` vgif:
            Pass `True` to create a video GIF.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :(raises): Similar to :meth:`.create_text_post`.

        :raises redditwarp.exceptions.RedditError:
            + `MISSING_VIDEO_URLS`:
                The `thumbnail` parameter was empty.
            + `NO_VIDEOS`:
                The subreddit does not accept video posts.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('kind', 'video' + ('gif' if vgif else ''))
            yield ('sr', sr)
            yield ('title', title)
            yield ('url', link)
            yield ('video_poster_url', thumbnail)
            yield ('sendreplies', '01'[reply_notifications])
            if spoiler: yield ('spoiler', '1')
            if nsfw: yield ('nsfw', '1')
            if oc: yield ('original_content', '1')
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        await self._client.request('POST', '/api/submit', data=dict(g()))

    async def create_gallery_post(self,
        sr: str,
        title: str,
        items: Sequence[GalleryItem],
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        oc: bool = False,
        collection_uuid: Optional[str] = None,
        flair_uuid: Optional[str] = None,
        flair_text: Optional[str] = None,
        event_start: Optional[str] = None,
        event_end: Optional[str] = None,
        event_tz: Optional[str] = None,
    ) -> int:
        """Create a gallery post.

        Behaves similarly to :meth:`.create_text_post`.

        .. .PARAMETERS

        :(parameters): Similar to :meth:`.create_text_post`.

        :param items:
            A list of gallery items.
        :type items: `Sequence`\\[:class:`~.dtos.submission.GalleryItem`]

        .. .RETURNS

        :(returns): Similar to :meth:`.create_text_post`.

        .. .RAISES

        :(raises): Similar to :meth:`.create_text_post`.
        """
        gallery_items: list[dict[str, str]] = [
            {
                'media_id': m.media_id,
                'caption': m.caption,
                'outbound_url': m.outbound_link,
            }
            for m in items
        ]

        def g() -> Iterable[tuple[str, JSON_ro]]:
            yield ('sr', sr)
            yield ('title', title)
            yield ('items', gallery_items)
            yield ('sendreplies', reply_notifications)
            if spoiler: yield ('spoiler', True)
            if nsfw: yield ('nsfw', True)
            if oc: yield ('original_content', True)
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        root = await self._client.request('POST', '/api/submit_gallery_post', json=dict(g()))
        return int(root['json']['data']['id'][3:], 36)

    async def create_poll_post(self,
        sr: str,
        title: str,
        body: str,
        options: Sequence[str],
        duration: int,
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        collection_uuid: Optional[str] = None,
        flair_uuid: Optional[str] = None,
        flair_text: Optional[str] = None,
        event_start: Optional[str] = None,
        event_end: Optional[str] = None,
        event_tz: Optional[str] = None,
    ) -> int:
        """Create a poll post.

        Behaves similarly to :meth:`.create_text_post`.

        .. .PARAMETERS

        :(parameters): Similar to :meth:`.create_text_post`.

        :param `str` body:
        :param `Sequence[str]` options:
        :param `int` duration:
            The number of days the poll runs for.

            Valid values are 1 to 7. If a number is specified outside
            this range it is clamped within range.

            The UI default is 3 days.

        .. .RETURNS

        :(returns): Similar to :meth:`.create_text_post`.

        .. .RAISES

        :(raises): Similar to :meth:`.create_text_post`.
        """
        def g() -> Iterable[tuple[str, JSON_ro]]:
            yield ('sr', sr)
            yield ('title', title)
            yield ('text', body)
            yield ('options', options)
            yield ('duration', duration)
            yield ('sendreplies', reply_notifications)
            if spoiler: yield ('spoiler', True)
            if nsfw: yield ('nsfw', True)
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        root = await self._client.request('POST', '/api/submit_poll_post', json=dict(g()))
        return int(root['json']['data']['id'][3:], 36)

    async def create_crosspost(self,
        sr: str,
        title: str,
        idn: int,
        *,
        reply_notifications: bool = True,
        spoiler: bool = False,
        nsfw: bool = False,
        oc: bool = False,
        collection_uuid: Optional[str] = None,
        flair_uuid: Optional[str] = None,
        flair_text: Optional[str] = None,
        event_start: Optional[str] = None,
        event_end: Optional[str] = None,
        event_tz: Optional[str] = None,
    ) -> int:
        """Create a crosspost.

        Behaves similarly to :meth:`.create_text_post`.

        .. .PARAMETERS

        :(parameters): Similar to :meth:`.create_text_post`.

        :param `int` idn:
            The ID of a submission.

        .. .RETURNS

        :(returns): Similar to :meth:`.create_text_post`.

        .. .RAISES

        :(raises): Similar to :meth:`.create_text_post`.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('kind', 'self')
            yield ('sr', sr)
            yield ('title', title)
            yield ('crosspost_parent', 't3_' + to_base36(idn))
            yield ('sendreplies', '01'[reply_notifications])
            if spoiler: yield ('spoiler', '1')
            if nsfw: yield ('nsfw', '1')
            if oc: yield ('original_content', '1')
            if collection_uuid: yield ('collection_id', collection_uuid)
            if flair_uuid: yield ('flair_id', flair_uuid)
            if flair_text: yield ('flair_text', flair_text)
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        root = await self._client.request('POST', '/api/submit', data=dict(g()))
        return int(root['json']['data']['id'], 36)

    async def edit_text_post_body(self, idn: int, body: Union[str, Mapping[str, JSON_ro]]) -> TextPost:
        """Edit the body text of a text post.

        .. .PARAMETERS

        :param `int` idn:
        :param body:
            Either markdown or richtext.
        :type body: `Union`\\[`str`, `Mapping`\\[`str`, :class:`~.types.JSON_ro`]]

        .. .RETURNS

        :rtype: :class:`~.models.submission_ASYNC.TextPost`

        .. .RAISES

        :raises:
            Same as comment :meth:`~.siteprocs.comment.SYNC.CommentProcedures.edit_body`
            but for submissions.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('thing_id', 't3_' + to_base36(idn))
            yield ('return_rtjson', '1')
            if isinstance(body, str):
                yield ('text', body)
            else:
                yield ('richtext_json', json.dumps(body))

        result = await self._client.request('POST', '/api/editusertext', files=dict(g()))
        return load_text_post(result, self._client)

    async def delete(self, idn: int) -> None:
        """Delete a submission.

        If the target doesn't exist or isn't valid, nothing happens.

        When a submission is deleted it's text content (if a text post)
        will be set to "`[deleted]`" and the submission will be unlisted
        from its subreddit. Users can still otherwise view and reply to
        deleted to submissions if they have a direct link to it.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        data = {'id': 't3_' + to_base36(idn)}
        await self._client.request('POST', '/api/del', data=data)

    async def lock(self, idn: int) -> None:
        """Lock a submission.

        Nothing happens if the target is already locked.

        .. hint::
           Locking prevents a submission/comment from receiving new comments.
           A locked submission is unable to receive any new comments.
           Locking a comment only stops direct comments, but
           existing child comments can still receive replies.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                The target doesn't exist or you don't have permission to lock it.
        """
        data = {'id': 't3_' + to_base36(idn)}
        await self._client.request('POST', '/api/lock', data=data)

    async def unlock(self, idn: int) -> None:
        """Unlock a submission.

        Behaves similarly to :meth:`.lock`.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises:
            Similar to :meth:`.lock`.
        """
        data = {'id': 't3_' + to_base36(idn)}
        await self._client.request('POST', '/api/unlock', data=data)

    async def vote(self, idn: int, direction: int) -> None:
        """Cast a vote on a submission.

        .. .PARAMETERS

        :param `int` idn:
        :param `int` direction:
            Either: `1` (upvote), `0` unvote, `-1` downvote.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The target could not be found.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'dir': str(direction),
        }
        await self._client.request('POST', '/api/vote', data=data)

    async def save(self, idn: int, category: Optional[str] = None) -> None:
        """Save a submission.

        .. .PARAMETERS

        :param `int` idn:
        :param `Optional[str]` category:
            A category/label.

            Requires Reddit Premium. Ignored if no Reddit Premium.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `404`:
                The category name specified was invalid.
        """
        data = {
            'id': 't3_' + to_base36(idn),
        }
        if category is not None:
            data['category'] = category
        await self._client.request('POST', '/api/save', data=data)

    async def unsave(self, idn: int) -> None:
        """Save a submission.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        data = {'id': 't3_' + to_base36(idn)}
        await self._client.request('POST', '/api/unsave', data=data)

    async def hide(self, idn: int) -> None:
        """Hide a submission.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `400`:
                The target was not found.
        """
        data = {'id': 't3_' + to_base36(idn)}
        await self._client.request('POST', '/api/hide', data=data)

    async def unhide(self, idn: int) -> None:
        """Unhide a submission.

        See :meth:`.hide`.
        """
        data = {'id': 't3_' + to_base36(idn)}
        await self._client.request('POST', '/api/unhide', data=data)

    def bulk_hide(self, ids: Iterable[int]) -> CallChunkCallingAsyncIterator[None]:
        """Bulk hide submissions.

        If *any* of the list of submission IDs don't exist then the endpoint will
        return a HTTP 400 status error and none of the submissions will be hidden.
        This can be annoying since if the list is long it can be hard to tell which
        ID is the culprit.

        .. .PARAMETERS

        :param `Iterable[int]` ids:

        .. .RETURNS

        :rtype: :class:`~.iterators.call_chunk_calling_async_iterator.CallChunkCallingAsyncIterator`\\[`None`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `400`:
                If any of the IDs were not found.
        """
        async def mass_hide(ids: Sequence[int]) -> None:
            id36s = map(to_base36, ids)
            full_id36s = map('t3_'.__add__, id36s)
            ids_str = ','.join(full_id36s)
            await self._client.request('POST', '/api/hide', data={'id': ids_str})

        return CallChunkCallingAsyncIterator(AsyncCallChunk(mass_hide, chunk) for chunk in chunked(ids, 300))

    def bulk_unhide(self, ids: Iterable[int]) -> CallChunkCallingAsyncIterator[None]:
        """Bulk hide submissions.

        See :meth:`.bulk_hide`.
        """
        async def mass_unhide(ids: Sequence[int]) -> None:
            id36s = map(to_base36, ids)
            full_id36s = map('t3_'.__add__, id36s)
            ids_str = ','.join(full_id36s)
            await self._client.request('POST', '/api/unhide', data={'id': ids_str})

        return CallChunkCallingAsyncIterator(AsyncCallChunk(mass_unhide, chunk) for chunk in chunked(ids, 300))

    async def mark_nsfw(self, idn: int) -> None:
        """Mark a submission as NSFW.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission to mark the target.
        """
        data = {'id': 't3_' + to_base36(idn)}
        await self._client.request('POST', '/api/marknsfw', data=data)

    async def unmark_nsfw(self, idn: int) -> None:
        """Unmark a submission as NSFW.

        See :meth:`.mark_nsfw`.
        """
        data = {'id': 't3_' + to_base36(idn)}
        await self._client.request('POST', '/api/unmarknsfw', data=data)

    async def mark_spoiler(self, idn: int) -> None:
        """Mark a submission as spoiler.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission to mark the target.
        """
        data = {'id': 't3_' + to_base36(idn)}
        await self._client.request('POST', '/api/spoiler', data=data)

    async def unmark_spoiler(self, idn: int) -> None:
        """Unmark a submission as spoiler.

        See :meth:`.mark_spoiler`.
        """
        data = {'id': 't3_' + to_base36(idn)}
        await self._client.request('POST', '/api/unspoiler', data=data)

    async def distinguish(self, idn: int) -> Submission:
        """Distinguish a submission.

        .. hint::

           Distinguishing decoratates the author's name by
           giving it a different color and putting a sigil beside it.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :returns: The target submission.
        :rtype: :class:`~.models.submission_ASYNC.Submission`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission to distinguish the target.
            + `404`:
                The target could not be found.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'how': 'yes',
        }
        root = await self._client.request('POST', '/api/distinguish', data=data)
        return load_submission(root['json']['data']['things'][0]['data'], self._client)

    async def undistinguish(self, idn: int) -> Submission:
        """Undistinguish a submission.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :returns: The target submission.
        :rtype: :class:`~.models.submission_ASYNC.Submission`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission.
            + `404`:
                The target could not be found.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'how': 'no',
        }
        root = await self._client.request('POST', '/api/distinguish', data=data)
        return load_submission(root['json']['data']['things'][0]['data'], self._client)

    async def sticky(self, idn: int, slot: Optional[int] = None) -> None:
        """Set a submission as sticky in its subreddit.

        .. hint::
           Stickied posts are shown at the top of the subreddit in the default 'Hot' listing.

        In a subreddit, there can be at most 2 sticked posts at a time.

        When stickying the `slot` parameter indicates which of the two positions the
        new post should occupy. If there is a sticked post in the slot specified by `slot`,
        it will be replaced. Otherwise the post will be placed in the bottom-most slot.
        If the number specified by `slot` is outside the valid range it will be clamped within range.

        Stickying a post that is already stickied causes a 409 (Conflict) HTTP error.
        Unstickying a post that isn't stickied does nothing.

        .. note::
           You cannot reorder sticky posts directly. You must unsticky and re-sticky them.

        .. .PARAMETERS

        :param `int` idn:
        :param `Optional[int]` slot:
            Which sticky slot to use.

            If not specified, the bottom-most slot will be used if available.
            If the sticky list is at maximum length, the bottom-most slot will
            be replaced with the new post.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission to sticky that post.
            + `409`:
                The post is already stickied.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'state': '1',
        }
        if slot is not None:
            data['num'] = str(slot)
        await self._client.request('POST', '/api/set_subreddit_sticky', data=data)

    async def unsticky(self, idn: int) -> None:
        """Unsticky a submission.

        See :meth:`.sticky`.

        Unstickying a post that isn't stickied does nothing.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'state': '0',
        }
        await self._client.request('POST', '/api/set_subreddit_sticky', data=data)

    async def pin_to_profile(self, idn: int, slot: Optional[int] = None) -> None:
        """Pin a post you created to your user profile.

        .. hint::
           Pinned posts show up at the start of the
           'Overview', or 'Submitted' (old UI) / 'POSTS' (redesign UI)
           user profile listings.

        A user can have at most 4 pinned posts at a time.

        The rules for the num parameter are the same as in subreddit stickying.
        See :meth:`.sticky`. However, there are differences when the slot number
        is unspecified. See the `slot` parameter description.

        .. note::
           This feature uses the same endpoint as :meth:`.sticky` but there are
           stark differences in insertion behaviour when `slot` is not specified.

           To summarise:

           * When subreddit stickying: the post will be placed at the **bottom** of the list.
             If the list is full then the bottom-most post will be **replaced**.
           * When user profile pinning: the post will be placed at the **top** of the list.
             If the list is full then the bottom-most post will be **evicted**.

        Pinning a post that is already pinned causes a 409 (Conflict) HTTP error.
        Unpinning a post that isn't pinned does nothing.

        .. note::
           You cannot reorder pinned posts directly. You must unpin and re-pin them.

        .. .PARAMETERS

        :param `int` idn:
        :param `Optional[int]` slot:
            Which pin slot to use.

            If `slot` is not specified, the new post is inserted at the top of the list.
            If the list is at maximum length, the least recently pinned post will be evicted.
            It acts like a queue.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You do not have permission to pin that post.
            + `409`:
                The post is already pinned.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'to_profile': '1',
            'state': '1',
        }
        if slot is not None:
            data['num'] = str(slot)
        await self._client.request('POST', '/api/set_subreddit_sticky', data=data)

    async def unpin_from_profile(self, idn: int) -> None:
        """Unpin a submission from your user profile.

        See :meth:`.pin_to_profile`.

        Unpinning a post that isn't pinned does nothing.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'to_profile': '1',
            'state': '0',
        }
        await self._client.request('POST', '/api/set_subreddit_sticky', data=data)

    async def set_contest_mode(self, idn: int, state: bool) -> None:
        """Set or unset 'contest mode' for a submission's comments.

        .. hint::

            In contest mode, vote counts are hidden and comments are displayed
            in a random order.

        .. .PARAMETERS

        :param `int` idn:
        :param `bool` state:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The specified ID was not found.
               - You do not have permission to modify the target.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'state': '01'[state],
        }
        await self._client.request('POST', '/api/set_contest_mode', data=data)

    async def set_suggested_sort(self, idn: int, sort: str) -> None:
        """Set or unset the suggested sort for a submission's comments.

        .. hint::

           When set, users will see comments in the suggested sort order by default.
           They can still manually change back to their preferred sort if they choose.

        .. .PARAMETERS

        :param `int` idn:
        :param `str` sort:
            Either: `confidence`, `top`, `new`, `controversial`,
            `old`, `random`, `qa`, `live`, `blank`.

            If not specified or an unknown value, the suggested sort will be unset.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The specified ID was not found.
               - You do not have permission to modify the target.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'sort': sort,
        }
        await self._client.request('POST', '/api/set_suggested_sort', data=data)

    async def enable_reply_notifications(self, idn: int) -> None:
        """Enable inbox reply notifications for a submission.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'state': '1'
        }
        await self._client.request('POST', '/api/sendreplies', data=data)

    async def disable_reply_notifications(self, idn: int) -> None:
        """Disable inbox reply notifications for a submission.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'state': '0'
        }
        await self._client.request('POST', '/api/sendreplies', data=data)

    async def set_event_time(self, idn: int,
            event_start: Optional[str] = None,
            event_end: Optional[str] = None,
            event_tz: Optional[str] = None) -> None:
        """Add or modify event times on a submission.

        Specify only `event_start` to change only the starting date.
        The same cannot be done for `event_end`, a 500 HTTP error will occur.

        If both `event_start` and `event_end` are specified then the `event_start` must be before
        `event_end` in time, otherwise a `MIN_EVENT_TIME` API error is returned.
        It is possible however to make a second request specifying only `event_start` to modify
        the start date so that `event_start` is after `event_end`. If this happens then the time
        difference can be longer than 7 days.

        .. .PARAMETERS

        :param `int` idn:
        :param `Optional[str]` event_start:
            A datetime ISO 8601 string. E.g. `2018-09-11T12:00:00`.
        :param `Optional[str]` event_end:
            A datetime ISO 8601 string.
        :param `Optional[str]` event_tz:
            A timezone. E.g., `America/Los_Angeles`.

            If not specified, effectively defaults to `UTC`.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `BAD_TIME`:
               - The value specified for `event_start` or `event_end` is in a bad format.
               - The date string specified for `event_start` or `event_end` is in the past.

               Note that this error will always indicate `event_start` is wrong even if
               it's `event_end` that needs fixing.
            + `MIN_EVENT_TIME`:
                The value specified for `event_tz` is invalid.
            + `MAX_EVENT_TIME`:
                The event can't be longer than 7 days.
            + `INVALID_TIMEZONE`:
                The value specified for `event_tz` is invalid.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `500`:
                The `event_start` parameter was not specified.
        """
        def g() -> Iterable[tuple[str, str]]:
            yield ('id', 't3_' + to_base36(idn))
            if event_start: yield ('event_start', event_start)
            if event_end: yield ('event_end', event_end)
            if event_tz: yield ('event_tz', event_tz)

        await self._client.request('POST', '/api/event_post_time', data=dict(g()))

    async def follow_event(self, idn: int) -> None:
        """Follow a post event.

        .. hint::
           Followers receive a push notification when the event starts.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                The target submission is not an event.
            + `404`:
                The target submission does not exist.
        """
        data = {
            'fullname': 't3_' + to_base36(idn),
            'follow': '1',
        }
        await self._client.request('POST', '/api/follow_post', data=data)

    async def unfollow_event(self, idn: int) -> None:
        """Unfollow a post event.

        See :meth:`.follow_event`.
        """
        data = {
            'fullname': 't3_' + to_base36(idn),
            'follow': '0',
        }
        await self._client.request('POST', '/api/follow_post', data=data)

    async def approve(self, idn: int) -> None:
        """Approve a submission.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               You do not have permission.

               - The target specified does not exist.
               - The target belongs to a subreddit you do not have permission over.
        """
        data = {'id': 't3_' + to_base36(idn)}
        await self._client.request('POST', '/api/approve', data=data)

    async def remove(self, idn: int) -> None:
        """Remove a submission.

        This is a moderator action.

        .. .PARAMETERS

        :param `int` idn:

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               You do not have permission.

               - The target specified does not exist.
               - The target belongs to a subreddit you do not have permission over.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'spam': '0',
        }
        await self._client.request('POST', '/api/remove', data=data)

    async def remove_spam(self, idn: int) -> None:
        """Remove as spam.

        See :meth:`.remove`.
        """
        data = {
            'id': 't3_' + to_base36(idn),
            'spam': '1',
        }
        await self._client.request('POST', '/api/remove', data=data)

    async def ignore_reports(self, idn: int) -> None:
        """Ignore reports on a submission.

        .. hint::
           If you ignore reports, you won't receive notifications and the
           ignored thing will be absent from moderation listings.

        Nothing happens if the target is already ignored.

        .. .PARAMETERS

        :param `int` idn:
            Submission ID.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The target specified does not exist.
               - The target belongs to a subreddit you do not have permission over.
        """
        await self._client.request('POST', '/api/ignore_reports', data={'id': 't3_' + to_base36(idn)})

    async def unignore_reports(self, idn: int) -> None:
        """Unignore reports on a submission.

        Nothing happens if the target is already unignored.

        .. .PARAMETERS

        :param `int` idn:
            Submission ID.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The target specified does not exist.
               - The target belongs to a subreddit you do not have permission over.
        """
        await self._client.request('POST', '/api/unignore_reports', data={'id': 't3_' + to_base36(idn)})

    async def snooze_reports(self, idn: int, reason: str) -> None:
        """Ignore a custom report reason in a subreddit for 7 days.

        .. .PARAMETERS

        :param `int` idn:
            Submission ID.
        :param `str` reason:
            The custom report reason to snooze.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The target specified does not exist.
               - The target belongs to a subreddit you do not have permission over.
        """
        data = {'id': 't3_' + to_base36(idn), 'reason': reason}
        await self._client.request('POST', '/api/snooze_reports', data=data)

    async def unsnooze_reports(self, idn: int, reason: str) -> None:
        """Unsnooze a custom report.

        See :meth:`.snooze_reports`.
        """
        data = {'id': 't3_' + to_base36(idn), 'reason': reason}
        await self._client.request('POST', '/api/unsnooze_reports', data=data)

    async def apply_removal_reason(self,
            idn: int,
            reason_id: Optional[str] = None,
            note: Optional[str] = None) -> None:
        """Set a removal reason on a removed submission.

        If the target is not a removed submission, nothing happens.

        .. .PARAMETERS

        :param `int` idn:
            Submission ID.
        :param `Optional[int]` reason_id:
            A removal reason ID.
        :param `Optional[str]` note:
            A note.

        .. .RETURNS

        :rtype: `None`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `JSON_MISSING_KEY`:
                Empty strings or null values were specified for both
                `reason_id` and `note` at the same time.
            + `NO_THING_ID`:
                The given target ID is not valid.
            + `INVALID_ID`:
                The reason ID is invalid.
            + `MUST_BE_PRESENT`:
                The subreddit specified does not exist.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The target specified does not belong to a subreddit you moderate.
        """
        target = 't3_' + to_base36(idn)
        json_data = {'item_ids': [target], 'reason_id': reason_id, 'mod_note': note}
        await self._client.request('POST', '/api/v1/modactions/removal_reasons', json=json_data)

    async def send_removal_comment(self,
            idn: int,
            title: str,
            message: str,
            *,
            exposed: bool = False,
            locked: bool = False) -> Comment:
        """Send a removal comment.

        Sends a removal reason comment to a user for a removed submission of theirs.

        This action can be performed multiple times.
        (The UI does not normally let you do this.)

        Unlike :meth:`.apply_removal_reason`, the target you specify must be a
        removed item otherwise an `INVALID_ID` API error is produced.

        .. .PARAMETERS

        :param `int` idn:
            ID of a removed submission.
        :param `str` title:
            A title.

            This is ultimately unused for removal comments, but a non-empty
            string must be specified or you'll get a `NO_TEXT` API error.

            The UI sends the title of the selected removal reason.
        :param `str` message:
            A message for the comment body.

            Can be an empty string. This is interesting because you can't
            normally create comments with empty bodies.
        :param `bool` exposed:
            If false (default), the comment will be created by a special
            moderator named `u/{subreddit}_ModTeam`.

            If true, the comment is created by the current user.
        :param `bool` locked:
            Lock the newly created comment.

        .. .RETURNS

        :returns: The newly created comment.
        :rtype: :class:`~.models.comment_ASYNC.Comment`

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `USER_REQUIRED`:
                There is no user context.
            + `NO_TEXT`:
                The value for the `title` parameter was empty.
            + `INVALID_ID`:
               - The target specified doesn't exist or is invalid.
               - The target specified is not a removed item.

            + `MUST_BE_PRESENT`:
                The subreddit specified does not exist.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
               - The target specified does not belong to a subreddit you moderate.
        """
        target = 't3_' + to_base36(idn)
        json_data = {
            'type': 'public' + ('' if exposed else '_as_subreddit'),
            'item_id': [target],
            'title': title,
            'message': message,
            'lock_comment': '01'[locked],
        }
        root = await self._client.request('POST', '/api/v1/modactions/removal_link_message', json=json_data)
        return load_comment(root, self._client)

    async def send_removal_message(self,
            idn: int,
            title: str,
            message: str,
            *,
            exposed: bool = False) -> None:
        """Send a removal message.

        Behaves similarly to :meth:`.send_removal_comment`.

        .. .PARAMETERS

        :param `int` idn:
            ID of a removed comment.
        :param `str` title:
            A title.

            A non-empty string must be specified or you'll get a `NO_TEXT` API error.

            The UI sends the title of the selected removal reason.
        :param `str` message:
            A message for the comment body.

            Can be an empty string.
        :param `bool` exposed:
            If false (default), the comment will be send on behalf of the subreddit.

            If true, the comment is sent by the current user.

        .. .RETURNS

        :returns: `None`

        .. .RAISES

        :(raises): See :meth:`.send_removal_comment`.
        """
        target = 't3_' + to_base36(idn)
        json_data = {
            'type': 'private' + ('_exposed' if exposed else ''),
            'item_id': [target],
            'title': title,
            'message': message,
        }
        await self._client.request('POST', '/api/v1/modactions/removal_link_message', json=json_data)

    def search(self, sr: str, query: str, amount: Optional[int] = None, *,
        sort: str = 'relevance', time: str = 'all',
    ) -> ImpartedPaginatorChainingAsyncIterator[SubmissionSearchAsyncPaginator, Submission]:
        """Search for submissions.

        .. .PARAMETERS

        :param `str` sr:
            A subreddit name.

            Use an empty string to search all of Reddit.
        :param `str` query:
            A search query.
        :param `Optional[int]` amount:
            The number of items to retrieve.
        :param `str` sort:
            Either: `relevance`, `hot`, `top`, `new`, or `comments`.

            Default: `relevance`.
        :param `str` time:
            Either: `all`, `hour`, `day`, `week`, `month`, `year`.

            Default: `all`.

        .. .RETURNS

        :rtype: :class:`~.pagination.paginator_chaining_async_iterator.ImpartedPaginatorChainingAsyncIterator`\\[:class:`~.pagination.paginators.submission_async1.SubmissionSearchAsyncPaginator`, :class:`~.models.submission_ASYNC.Submission`]
        """
        url = '/search'
        if sr:
            url = f'/r/{sr}/search'
        p = SubmissionSearchAsyncPaginator(
                self._client, url,
                params={'q': query, 'restrict_sr': '1'},
                    sort=sort, time=time)
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def duplicates(self, target: int, amount: Optional[int] = None, *,
        sort: str = 'num_comments',
    ) -> ImpartedPaginatorChainingAsyncIterator[SubmissionDuplicatesAsyncPaginator, Submission]:
        """Get crossposts for a submission.

        .. .PARAMETERS

        :param `int` target:
            Submission ID.
        :param `Optional[int]` amount:
            The number of items to retrieve.
        :param `str` sort:
            Either `num_comments` or `new`.

            Default: `num_comments`.

        .. .RETURNS

        :rtype: :class:`~.pagination.paginator_chaining_async_iterator.ImpartedPaginatorChainingAsyncIterator`\\[:class:`~.pagination.paginators.submission_async1.SubmissionDuplicatesAsyncPaginator`, :class:`~.models.submission_ASYNC.Submission`]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                Fetching some submissions resulted in a 403.
            + `404`:
                The target submission could not be found.
        """
        subm_id = to_base36(target)
        p = SubmissionDuplicatesAsyncPaginator(self._client, f'/duplicates/{subm_id}', sort=sort)
        return ImpartedPaginatorChainingAsyncIterator(p, amount)




from __future__ import annotations
from typing import Sequence, Optional

import sys
import os.path as op
from os import getenv
from configparser import ConfigParser

def _safe_getenv(key: str) -> str:
    return getenv(key, '')

def get_praw_ini_potential_locations() -> Sequence[str]:
    """Return a list of potential `praw.ini` locations."""
    root_pkg_dirname = ''
    if __name__ != '__main__':
        root_pkg_name, _, _ = __name__.partition('.')
        root_pkg_module = sys.modules[root_pkg_name]
        root_pkg_dirname = op.dirname(root_pkg_module.__file__ or '')

    return [
        op.join(*path_components, 'praw.ini')
        for path_components in [
            [root_pkg_dirname],  # Package defaults
            [_safe_getenv('APPDATA')],  # Windows
            [_safe_getenv('HOME'), '.config'],  # Linux and macOS
            [_safe_getenv('XDG_CONFIG_HOME')],  # Linux
            ['.'],  # Current directory
        ]
        if path_components[0]
    ]

def get_praw_config(filepath: Optional[str] = None) -> ConfigParser:
    """Return a `ConfigParser` initialised with the standard `praw.ini` locations."""
    config = ConfigParser()
    config.read(get_praw_ini_potential_locations() if filepath is None else filepath)
    return config




from __future__ import annotations
from typing import TYPE_CHECKING, TypeVar, Generic, Optional
if TYPE_CHECKING:
    from typing import Iterable, Iterator, Callable, Awaitable, AsyncIterator

from .stubborn_caller_async_iterator import StubbornCallerAsyncIterator

T = TypeVar('T')

class CallChunkCallingAsyncIterator(Generic[T]):
    """Evaluate call chunks and return their results."""

    @property
    def current_callable(self) -> Optional[Callable[[], Awaitable[T]]]:
        return self.__calling_itr.current

    @current_callable.setter
    def current_callable(self, value: Optional[Callable[[], Awaitable[T]]]) -> None:
        self.__calling_itr.current = value

    def __init__(self, chunks: Iterable[Callable[[], Awaitable[T]]]) -> None:
        self.__chunking_itr: Iterator[Callable[[], Awaitable[T]]] = iter(chunks)
        self.__calling_itr: StubbornCallerAsyncIterator[T] = StubbornCallerAsyncIterator(self.__chunking_itr)

    def __aiter__(self) -> AsyncIterator[T]:
        return self

    async def __anext__(self) -> T:
        return await self.__calling_itr.__anext__()

    def get_chunking_iterator(self) -> Iterator[Callable[[], Awaitable[T]]]:
        return self.__chunking_itr




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any
if TYPE_CHECKING:
    from .types import ClientCredentials, AuthorizationGrant
    from ..http.http_client_SYNC import HTTPClient

from ..http.util.json_loading import load_json_from_response_but_prefer_status_code_exception_on_failure
from .token import Token
from .utils import apply_basic_auth
from .exceptions import raise_for_token_server_response_error

class TokenObtainmentClient:
    """Exchanges an authorisation grant for an OAuth2 token."""

    def __init__(self, http: HTTPClient, url: str,
            client_creds: ClientCredentials,
            grant: AuthorizationGrant) -> None:
        """
        .. .PARAMETERS

        :param http:
            An instance of `HTTPClient` class that makes the HTTP requests.
        :type http:
            :class:`~.http.http_client_SYNC.HTTPClient`
        :param url:
            The token obtainment endpoint URL to make the request to.
        :type url:
            str
        :param client_creds:
            The client credentials to use.
        :type client_creds:
            :data:`~.types.ClientCredentials`
        :param grant:
            An authorization grant.
        :type grant:
            :data:`~.types.AuthorizationGrant`
        """
        self.http: HTTPClient = http
        ("")
        self.url: str = url
        ("")
        self.client_creds: ClientCredentials = client_creds
        ("")
        self.grant: AuthorizationGrant = grant
        ("")

    def fetch_data(self) -> Mapping[str, Any]:
        """Exchange an authorisation grant for OAuth2 token data.

        Makes a POST request to the specified URL (:attr:`url`) with
        the provided client credentials (:attr:`client_creds`) and
        authorization grant (:attr:`grant`).

        .. .RETURNS

        :returns:
            A dictionary of OAuth2 token data.
        :rtype: Mapping[str, Any]

        .. .RAISES

        :raises redditwarp.http.exceptions.StatusCodeException:
            The response from the server is not a success status code.
        :raises ValueError:
            The response from the server cannot be parsed as JSON.
        """
        headers: dict[str, str] = {}
        apply_basic_auth(headers, *self.client_creds)
        resp = self.http.request('POST', self.url, data=self.grant, headers=headers)

        json_data = load_json_from_response_but_prefer_status_code_exception_on_failure(resp)
        raise_for_token_server_response_error(json_data)
        resp.ensure_successful_status()
        return json_data

    def fetch_token(self) -> Token:
        """Obtain a `Token` instance.

        Makes use of the :meth:`.fetch_data` method to obtain OAuth2 token
        data and then constructs a :class:`~redditwarp.auth.token.Token` instance from the obtained data.

        .. .RETURNS

        :returns:
            An instance of the `Token` class representing the obtained OAuth2 token.
        :rtype: :class:`~redditwarp.auth.token.Token`
        """
        return Token.from_dict(self.fetch_data())




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from collections.abc import Mapping

from dataclasses import dataclass


@dataclass(repr=False, eq=False)
class Request:
    verb: str
    url: str
    headers: Mapping[str, str]
    data: bytes

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} [{self.verb}]>"




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.subreddit_ASYNC import Subreddit

from ...pagination.paginator_chaining_async_iterator import ImpartedPaginatorChainingAsyncIterator
from ...pagination.paginators.user.async1 import UserSubredditListingAsyncPaginator

class PullSubreddits:
    def __init__(self, client: Client) -> None:
        self._client = client

    def popular(self, amount: Optional[int] = None
            ) -> ImpartedPaginatorChainingAsyncIterator[UserSubredditListingAsyncPaginator, Subreddit]:
        p = UserSubredditListingAsyncPaginator(self._client, '/users/popular')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)

    def new(self, amount: Optional[int] = None
            ) -> ImpartedPaginatorChainingAsyncIterator[UserSubredditListingAsyncPaginator, Subreddit]:
        p = UserSubredditListingAsyncPaginator(self._client, '/users/new')
        return ImpartedPaginatorChainingAsyncIterator(p, amount)




from __future__ import annotations
from typing import Any, Mapping, Optional

from functools import cached_property
from datetime import datetime, timezone

class SubredditUser:
    @cached_property
    def added_at(self) -> datetime:
        return datetime.fromtimestamp(self.added_ut, timezone.utc)

    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        full_id36: str = d['id']
        _, _, id36 = full_id36.partition('_')
        self.id36: str = id36
        ("""
            User ID as a base 36 number.
            """)
        self.id: int = int(id36, 36)
        ("""
            User ID.
            """)
        self.name: str = d['username']
        ("""
            Username.
            """)
        self.account_icon: str = d['accountIcon']
        ("""
            URL of the user account's icon.
            """)
        self.icon_size: tuple[int, int] = d['iconSize']
        ("""
            Usually `(256, 256)`.
            """)
        self.added_ut: int = 0
        ("")

class Moderator(SubredditUser):
    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.added_ut: int = d['moddedAtUTC']
        ("""
            UNIX timestamp of when the user added as a moderator.
            """)
        self.flair_text: str = d['authorFlairText']
        ("""
            The flair text of the moderator.
            """)
        self.permissions: set[str] = {k for k, v in d['modPermissions'].items() if v}
        ("""
            Values: `all`, `wiki`, `chat_operator`, `chat_config`,
            `posts`, `access`, `mail`, `config`, `flair`.
            """)
        self.post_karma: int = d['postKarma']
        ("")

class ApprovedUser(SubredditUser):
    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.added_ut: int = d['approvedAtUTC']
        ("""
            UNIX timestamp of when the user added as an approved user.
            """)

class BannedUser(SubredditUser):
    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.added_ut: int = d['bannedAtUTC']
        ("""
            UNIX timestamp of when the user was banned.
            """)
        self.agent_name: str = d['bannedBy']
        ("""
            Name of the moderator who banned the user.

            Unknown what happens if the user is deleted.
            Is the value `[deleted]`, an empty string, or does the field in the
            underlying object not exist?
            If you have any information about this, please open an issue report at
            `<https://github.com/Pyprohly/redditwarp/issues>`_.
            """)
        self.reason: str = d['reason'] or ''
        ("""
            Ban reason.
            """)
        self.note: str = d['modNote'] or ''
        ("""
            A moderator note.
            """)
        self.message: str = d['banMessage']
        ("""
            The message that was sent to the user when they were banned.
            """)
        self.days_remaining: Optional[int] = d['duration']
        ("""
            Number of days until the ban is lifted.

            Value `None` if the ban is permanent.
            """)

class MutedUser(SubredditUser):
    def __init__(self, d: Mapping[str, Any]) -> None:
        super().__init__(d)
        self.added_ut: int = d['mutedAtUTC']
        ("""
            UNIX timestamp of when the user was muted.
            """)
        self.agent_name: str = d['mutedBy']
        ("""
            Name of the moderator who muted the user.

            Unknown what happens if the user is deleted.
            Is the value `[deleted]`, an empty string, or does the field in the
            underlying object not exist?
            If you have any information about this, please open an issue report at
            `<https://github.com/Pyprohly/redditwarp/issues>`_.
            """)
        self.reason: str = d['reason']
        ("""
            A moderator note.
            """)




from . import http_client_ASYNC  # noqa: F401
from .http_client_ASYNC import HTTPClient as HTTPClient  # noqa: F401
from . import reddit_token_obtainment_client_ASYNC  # noqa: F401
from .reddit_token_obtainment_client_ASYNC import RedditTokenObtainmentClient as RedditTokenObtainmentClient  # noqa: F401
from . import authorizer_ASYNC  # noqa: F401
from .authorizer_ASYNC import Authorizer as Authorizer, Authorized as Authorized  # noqa: F401
from . import rate_limited_ASYNC  # noqa: F401
from .rate_limited_ASYNC import RateLimited as RateLimited  # noqa: F401




from __future__ import annotations
from typing import Any, Mapping, Sequence, Iterator, Union, overload

from dataclasses import dataclass

from .datamemento import DatamementoDataclassesMixin


@dataclass(repr=False, eq=False)
class FlairTemplate(DatamementoDataclassesMixin):
    d: Mapping[str, Any]
    uuid: str
    text_mode: str
    ("""
        Either `text` or `richtext`.
        """)
    text: str
    css_class: str
    bg_color: str
    ("""
        A background color hex string. E.g., `#46d160`.

        If a user flair template and the background color is unset then
        the value is an empty string. The value cannot be empty for
        post flair templates.
        """)
    fg_color_scheme: str
    ("""
        Either `dark` or `light`.
        """)
    mod_only: bool
    ("""
        Whether flair is only available for mods to select.
        """)
    text_editable: bool
    ("""
        Whether users are able to edit the flair text.
        """)
    allowable_content: str
    ("""
        Either: `all`, `emoji`, `text`.

        Value `all` if both text and emojis are allowed in the flair text,
        `emoji` if only emojis are allowed, `text` if only text is allowed.
        """)
    max_emojis: int
    ("""
        Maximum number of emojis allowed in a flair.

        An integer from 1 to 10.
        """)


@dataclass(repr=False, eq=False)
class FlairTemplateChoice:
    uuid: str
    text: str
    css_class: str
    text_editable: bool

@dataclass(repr=False, eq=False)
class FlairTemplateChoices(Sequence[FlairTemplateChoice]):
    choices: Sequence[FlairTemplateChoice]
    subreddit_flair_user_position: str
    ("""
        Either: `left`, `right`, or empty string.
        """)

    def __len__(self) -> int:
        return len(self.choices)

    def __contains__(self, item: object) -> bool:
        return item in self.choices

    def __iter__(self) -> Iterator[FlairTemplateChoice]:
        return iter(self.choices)

    @overload
    def __getitem__(self, index: int) -> FlairTemplateChoice: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[FlairTemplateChoice]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[FlairTemplateChoice, Sequence[FlairTemplateChoice]]:
        return self.choices[index]


@dataclass(repr=False, eq=False)
class UserFlairAssociation:
    user: str
    text: str
    css_class: str
    has_had_flair_css_class_assigned_before_in_subreddit_when_no_flair_template_assigned: bool



"""Specialised classes and functions.

This subpackage exists mainly so that the top-level namespace is not cluttered by boring code.
"""

from . import const  # noqa: F401
from . import exceptions  # noqa: F401
from . import grants  # noqa: F401




from __future__ import annotations
from typing import Any, Mapping

from ..models.subreddit_user import (
    Moderator,
    ApprovedUser,
    BannedUser,
    MutedUser,
)

def load_moderator(d: Mapping[str, Any]) -> Moderator:
    return Moderator(d)

def load_approved_user(d: Mapping[str, Any]) -> ApprovedUser:
    return ApprovedUser(d)

def load_banned_user(d: Mapping[str, Any]) -> BannedUser:
    return BannedUser(d)

def load_muted_user(d: Mapping[str, Any]) -> MutedUser:
    return MutedUser(d)




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Mapping, Optional, Sequence, Iterator, overload, Union, Any
if TYPE_CHECKING:
    from ...client_SYNC import Client

from ...models.submission_SYNC import Submission
from .listing.mixins.time_SYNC import Time
from .listing.mixins.sort_SYNC import Sort
from .listing.submission_listing_paginator import SubmissionListingPaginator
from .listing.listing_paginator import ListingPaginator
from ...model_loaders.submission_SYNC import load_submission

class SubmissionSearchPaginator(
    Time[Submission],
    Sort[Submission],
    SubmissionListingPaginator,
):
    def __init__(self,
        client: Client,
        url: str,
        *,
        params: Optional[Mapping[str, str]] = None,
        sort: str = 'relevance',
        time: str = 'all',
    ) -> None:
        super().__init__(client, url, params=params)
        self.sort: str = sort
        ("")
        self.time: str = time
        ("")


class SubmissionDuplicates(Sequence[Submission]):
    def __init__(self, dups: Sequence[Submission], origin: Submission) -> None:
        self.origin: Submission = origin
        ("")
        self.dups: Sequence[Submission] = dups
        ("")

    def __len__(self) -> int:
        return len(self.dups)

    def __contains__(self, item: object) -> bool:
        return item in self.dups

    def __iter__(self) -> Iterator[Submission]:
        return iter(self.dups)

    @overload
    def __getitem__(self, index: int) -> Submission: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[Submission]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[Submission, Sequence[Submission]]:
        return self.dups[index]

class SubmissionDuplicatesPaginator(ListingPaginator[Submission]):
    def __init__(self,
        client: Client,
        url: str,
        *,
        sort: str = 'num_comments',
        crossposts_only: bool = False,
    ) -> None:
        super().__init__(client, url)
        self.sort: str = sort
        ("")
        self.crossposts_only: bool = crossposts_only
        ("")

    def _generate_params(self) -> Iterable[tuple[str, str]]:
        yield from super()._generate_params()
        if self.crossposts_only:
            yield ('crossposts_only', '1')
        if self.sort:
            yield ('sort', self.sort)

    def _fetch_data(self) -> Any:
        params = dict(self._generate_params())
        root = self.client.request('GET', self.url, params=params)
        data = root[1]['data']
        children = data['children']

        dist: int = x if (x := data['dist']) else len(children)
        self.after_count += (1 if self.direction else -1) * dist
        self.before_count: int = self.after_count - dist + 1

        suggested_forward_cursor = data['after'] or ''
        suggested_backward_cursor = data['before'] or ''
        if children:
            self.after: str = suggested_forward_cursor if suggested_forward_cursor else self.cursor_extractor(children[-1])
            self.before: str = suggested_backward_cursor if suggested_backward_cursor else self.cursor_extractor(children[0])
        self.has_after: bool = bool(suggested_forward_cursor)
        self.has_before: bool = bool(suggested_backward_cursor)

        return root

    def fetch(self) -> SubmissionDuplicates:
        root = self._fetch_data()
        origin = load_submission(root[0]['data']['children'][0]['data'], self.client)
        data = root[1]['data']
        return SubmissionDuplicates([load_submission(d['data'], self.client) for d in data['children']], origin)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.subreddit_SYNC import Subreddit

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.account_sync1 import (
    SubscribedListingPaginator,
    ContributingListingPaginator,
    ModeratingListingPaginator,
)

class PullSubreddits:
    def __init__(self, client: Client) -> None:
        self._client = client

    def subscribed(self, amount: Optional[int] = None
            ) -> ImpartedPaginatorChainingIterator[SubscribedListingPaginator, Subreddit]:
        """Pull subreddits the current user is subscribed to."""
        p = SubscribedListingPaginator(self._client, '/subreddits/mine/subscriber')
        return ImpartedPaginatorChainingIterator(p, amount)

    def contributing(self, amount: Optional[int] = None
            ) -> ImpartedPaginatorChainingIterator[ContributingListingPaginator, Subreddit]:
        """Pull subreddits the current user is an approved user in."""
        p = ContributingListingPaginator(self._client, '/subreddits/mine/contributor')
        return ImpartedPaginatorChainingIterator(p, amount)

    def moderating(self, amount: Optional[int] = None
            ) -> ImpartedPaginatorChainingIterator[ModeratingListingPaginator, Subreddit]:
        """Pull subreddits the current user is a moderator of."""
        p = ModeratingListingPaginator(self._client, '/subreddits/mine/moderator')
        return ImpartedPaginatorChainingIterator(p, amount)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.subreddit_user import (
        Moderator,
        ApprovedUser,
        BannedUser,
        MutedUser,
    )

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.moderation.sync1 import (
    ModeratorsPaginator,
    ApprovedUsersPaginator,
    BannedUsersPaginator,
    MutedUsersPaginator,
)

class PullUsers:
    def __init__(self, client: Client) -> None:
        self._client = client

    def moderators(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[ModeratorsPaginator, Moderator]:
        """
        Behaves similarly to :meth:`.approved_users`.
        """
        p = ModeratorsPaginator(self._client, f'/api/v1/{sr}/moderators')
        return ImpartedPaginatorChainingIterator(p, amount)

    def moderator_invitations(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[ModeratorsPaginator, Moderator]:
        """
        Behaves similarly to :meth:`.approved_users`.
        """
        p = ModeratorsPaginator(self._client, f'/api/v1/{sr}/moderators_invited')
        return ImpartedPaginatorChainingIterator(p, amount)

    def editable_moderators(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[ModeratorsPaginator, Moderator]:
        """
        Behaves similarly to :meth:`.approved_users`.
        """
        p = ModeratorsPaginator(self._client, f'/api/v1/{sr}/moderators_editable')
        return ImpartedPaginatorChainingIterator(p, amount)

    def approved_users(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[ApprovedUsersPaginator, ApprovedUser]:
        """Get redditors that relate to a subreddit.

        .. .PARAMETERS

        :param `str` sr:
        :param `Optional[int]` amount:

        .. .RETURNS

        :rtype: :class:`~.pagination.paginator_chaining_iterator.ImpartedPaginatorChainingIterator`\\[:class:`~.pagination.paginators.moderation.sync1.ApprovedUsersPaginator`, :class:`~.models.subreddit_user.ApprovedUser`]

        .. .RAISES

        :raises redditwarp.exceptions.RedditError:
            + `SUBREDDIT_NOEXIST`:
                The target subreddit does not exist.
            + `SUBREDDIT_NO_ACCESS`:
                The subreddit cannot be accessed.
        :raises redditwarp.http.exceptions.StatusCodeException:
            + `403`:
                You don't have permission.
        """
        p = ApprovedUsersPaginator(self._client, f'/api/v1/{sr}/contributors')
        return ImpartedPaginatorChainingIterator(p, amount)

    def banned(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[BannedUsersPaginator, BannedUser]:
        """
        Behaves similarly to :meth:`.approved_users`.
        """
        p = BannedUsersPaginator(self._client, f'/api/v1/{sr}/banned')
        return ImpartedPaginatorChainingIterator(p, amount)

    def muted(self, sr: str, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[MutedUsersPaginator, MutedUser]:
        """
        Behaves similarly to :meth:`.approved_users`.
        """
        p = MutedUsersPaginator(self._client, f'/api/v1/{sr}/muted')
        return ImpartedPaginatorChainingIterator(p, amount)




from __future__ import annotations
from typing import Optional, Union

import json

import urllib3  # type: ignore[import]
import urllib3.exceptions as urllib3_exceptions  # type: ignore[import]

from ... import exceptions
from ... import payload
from ...send_params import SendParams
from ...exchange import Exchange
from ...request import Request
from ...response import UResponse
from ..reg_SYNC import register
from ..connector_SYNC import Connector
from ...util.merge_query_params import merge_query_params


def _get_effective_timeout(v: float) -> float:
    if 0 > v != -1:
        raise ValueError(f"invalid `timeout` value: {v}")
    return v

def _get_effective_follow_redirects(v: Optional[bool]) -> bool:
    if v is None:
        raise ValueError(f"invalid `follow_redirects` value: {v}")
    return v


class Urllib3Connector(Connector):
    def __init__(self,
        http: urllib3.poolmanager.PoolManager,
    ) -> None:
        super().__init__()
        self.http: urllib3.poolmanager.PoolManager = http
        ("")

    def _send(self, p: SendParams) -> Exchange:
        r = p.requisition

        etv = _get_effective_timeout(p.timeout)
        tmo = None if etv == -1 else etv

        follow_redirects = _get_effective_follow_redirects(p.follow_redirects)

        url = merge_query_params(r.url, r.params)
        headers = dict(r.headers)
        pld = r.payload

        resp: urllib3.response.HTTPResponse
        try:
            if pld is None:
                resp = self.http.urlopen(
                    r.verb,
                    url,
                    headers=headers,
                    timeout=tmo,
                    redirect=follow_redirects,
                )

            elif isinstance(pld, payload.Bytes):
                headers['Content-Type'] = pld.get_media_type()
                resp = self.http.urlopen(
                    r.verb,
                    url,
                    headers=headers,
                    body=pld.data,
                    timeout=tmo,
                    redirect=follow_redirects,
                )

            elif isinstance(pld, payload.Text):
                headers['Content-Type'] = pld.get_media_type()
                resp = self.http.urlopen(
                    r.verb,
                    url,
                    headers=headers,
                    body=pld.text.encode(),
                    timeout=tmo,
                    redirect=follow_redirects,
                )

            elif isinstance(pld, payload.JSON):
                headers['Content-Type'] = pld.get_media_type()
                resp = self.http.urlopen(
                    r.verb,
                    url,
                    headers=headers,
                    body=json.dumps(pld.json).encode(),
                    timeout=tmo,
                    redirect=follow_redirects,
                )

            elif isinstance(pld, payload.URLEncodedFormData):
                fields0: dict[str, str] = dict(pld.data)
                resp = self.http.request_encode_body(
                    r.verb,
                    url,
                    headers=headers,
                    fields=fields0,
                    timeout=tmo,
                    redirect=follow_redirects,
                    encode_multipart=False,
                )

            elif isinstance(pld, payload.MultipartFormData):
                fields: dict[str, Union[str, tuple[Optional[str], Union[str, bytes], Optional[str]]]] = {}
                for pt in pld.parts:
                    if isinstance(pt, payload.MultipartFormData.TextField):
                        fields[pt.name] = pt.text
                    elif isinstance(pt, payload.MultipartFormData.FileField):
                        fields[pt.name] = (pt.filename, pt.file.read(), pt.content_type)
                    else:
                        raise ValueError('unexpected multipart field type: ' + repr(pt))

                resp = self.http.request_encode_body(
                    r.verb,
                    url,
                    headers=headers,
                    fields=fields,
                    timeout=tmo,
                    redirect=follow_redirects,
                )

            else:
                raise Exception(f"unsupported payload type: {pld.__class__.__name__!r}")

        except urllib3_exceptions.ReadTimeoutError as cause:
            raise exceptions.TimeoutException from cause
        except Exception as cause:
            raise exceptions.TransportError from cause

        x_requ = Request(
            verb=r.verb,
            url=url,
            headers={},
            data=b'',
        )
        x_resp = UResponse(
            status=resp.status,
            headers=resp.headers,
            data=resp.data,
            underlying_object=resp,
        )
        return Exchange(
            requisition=r,
            request=x_requ,
            response=x_resp,
            history=(),
        )

    def _close(self) -> None:
        self.http.clear()


def new_connector() -> Urllib3Connector:
    return Urllib3Connector(urllib3.PoolManager())


name: str = urllib3.__name__
version: str = urllib3.__version__
register(
    adaptor_module_name=__name__,
    name=name,
    version=version,
    new_connector=new_connector,
)




from __future__ import annotations
from typing import Any, Mapping

from dataclasses import dataclass

from .datamemento import DatamementoDataclassesMixin

@dataclass(repr=False, eq=False)
class Trophy(DatamementoDataclassesMixin):
    d: Mapping[str, Any]
    name: str
    ("""
        E.g., `Three-Year Club`.
        """)
    description: str
    icon_40: str
    ("""
        The URL of a 41x41 px icon for the trophy.

        E.g., `https://www.redditstatic.com/awards2/3_year_club-40.png`.
        """)
    icon_70: str
    ("""
        The URL of a 71x71 px icon for the trophy.

        E.g., `https://www.redditstatic.com/awards2/3_year_club-70.png`.
        """)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.submission_SYNC import Submission

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.front_sync1 import (
    BestListingPaginator,
    HotListingPaginator,
    NewListingPaginator,
    TopListingPaginator,
    RisingListingPaginator,
    ControversialListingPaginator,
    GildedListingPaginator,
)

class Pull:
    def __init__(self, client: Client) -> None:
        self._client = client

    def best(self, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[BestListingPaginator, Submission]:
        p = BestListingPaginator(self._client, '/best')
        return ImpartedPaginatorChainingIterator(p, amount)

    def hot(self, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[HotListingPaginator, Submission]:
        p = HotListingPaginator(self._client, '/hot')
        return ImpartedPaginatorChainingIterator(p, amount)

    def new(self, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[NewListingPaginator, Submission]:
        p = NewListingPaginator(self._client, '/new')
        return ImpartedPaginatorChainingIterator(p, amount)

    def top(self, amount: Optional[int] = None, *,
            time: str = 'day',
            ) -> ImpartedPaginatorChainingIterator[TopListingPaginator, Submission]:
        p = TopListingPaginator(self._client, '/top')
        p.time = time
        return ImpartedPaginatorChainingIterator(p, amount)

    def rising(self, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[RisingListingPaginator, Submission]:
        p = RisingListingPaginator(self._client, '/rising')
        return ImpartedPaginatorChainingIterator(p, amount)

    def controversial(self, amount: Optional[int] = None, *,
            time: str = 'day',
        ) -> ImpartedPaginatorChainingIterator[ControversialListingPaginator, Submission]:
        p = ControversialListingPaginator(self._client, '/controversial')
        p.time = time
        return ImpartedPaginatorChainingIterator(p, amount)

    def gilded(self, amount: Optional[int] = None) -> ImpartedPaginatorChainingIterator[GildedListingPaginator, object]:
        p = GildedListingPaginator(self._client, '/gilded')
        return ImpartedPaginatorChainingIterator(p, amount)




from __future__ import annotations
from typing import TYPE_CHECKING, Iterable, Optional
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.subreddit_ASYNC import Subreddit
    from ..stream_ASYNC import IStandardStreamEventSubject
    from ...pagination.async_paginator import CursorAsyncPaginator

from ..stream_ASYNC import Stream


def user_subreddit_stream_extractor(x: Subreddit) -> object:
    return x.id

def get_user_subreddit_stream_paginator(client: Client) -> CursorAsyncPaginator[Subreddit]:
    return client.p.user.pull_subreddits.new().get_paginator()

def make_user_subreddit_stream(paginator: CursorAsyncPaginator[Subreddit], past: Optional[Iterable[Subreddit]] = None, seen: Optional[Iterable[object]] = None) -> IStandardStreamEventSubject[Subreddit]:
    return Stream(paginator, user_subreddit_stream_extractor, past=past, seen=seen)

def create_user_subreddit_stream(client: Client) -> IStandardStreamEventSubject[Subreddit]:
    return make_user_subreddit_stream(get_user_subreddit_stream_paginator(client))




from __future__ import annotations
from typing import Mapping, Any, Sequence, Iterator, overload, Union

from enum import Enum, auto

class RuleScope(Enum):
    ALL = auto()
    SUBMISSION = auto()
    COMMENT = auto()

class Rule:
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.scope: RuleScope = {
            'all': RuleScope.ALL,
            'link': RuleScope.SUBMISSION,
            'comment': RuleScope.COMMENT,
        }[d['kind']]
        ("""
            Content this rule applies to.

            Enum: :class:`.RuleScope`.
            """)
        self.short_name: str = d['short_name']
        ("""
            Short description.

            Up to 100 characters.
            """)
        self.description: str = d['description']
        ("""
            Description.

            Up to 500 characters.
            """)
        self.description_html: str = d.get('description_html', '')
        ("""
            Same as :attr:`description` but HTML formatted.
            """)
        self.violation_reason: str = d['violation_reason']
        ("""
            Violation reason text.

            Up to 100 characters.

            Value matches `short_name` if left empty in the UI. It's unfortunately not possible
            to tell if this field is empty through the API.
            """)
        self.created_ut: int = int(d['created_utc'])
        ("""
            Unix timestamp of when the rule was created.
            """)


class SubredditRules(Sequence[Rule]):
    def __init__(self, d: Mapping[str, Any]) -> None:
        self.d: Mapping[str, Any] = d
        ("")
        self._rules = [Rule(o) for o in d['rules']]

    def __len__(self) -> int:
        return len(self._rules)

    def __contains__(self, item: object) -> bool:
        return item in self._rules

    def __iter__(self) -> Iterator[Rule]:
        return iter(self._rules)

    @overload
    def __getitem__(self, index: int) -> Rule: ...
    @overload
    def __getitem__(self, index: slice) -> Sequence[Rule]: ...
    def __getitem__(self, index: Union[int, slice]) -> Union[Rule, Sequence[Rule]]:
        return self._rules[index]




from __future__ import annotations
from typing import Mapping, Any

from ..models.subreddit_style_asset_upload_lease import SubredditStyleAssetUploadLease

def load_subreddit_style_asset_upload_lease(d: Mapping[str, Any]) -> SubredditStyleAssetUploadLease:
    lease_data = d['s3UploadLease']
    endpoint = f'https:{x}' if (x := lease_data['action']).startswith('//') else x
    fields = {field['name']: field['value'] for field in lease_data['fields']}
    s3_object_key = fields['key']
    return SubredditStyleAssetUploadLease(
        d=d,
        endpoint=endpoint,
        fields=fields,
        s3_object_key=s3_object_key,
        location=f"{endpoint}/{s3_object_key}",
    )




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_ASYNC import Client
    from ..models.submission_ASYNC import Submission

from ..models.submission_ASYNC import (
    LinkPost,
    TextPost,
    GalleryPost,
    PollPost,
    CrosspostSubmission,
)

def load_submission(d: Mapping[str, Any], client: Client) -> Submission:
    if d.get('is_gallery', False):
        return load_gallery_post(d, client)
    if 'poll_data' in d:
        return load_poll_post(d, client)
    if 'crosspost_parent' in d:
        return load_crosspost_submission(d, client)
    if d['is_self']:
        return load_text_post(d, client)
    if 'url_overridden_by_dest' in d:
        return load_link_post(d, client)
    raise Exception('unknown post type')


def load_text_post(d: Mapping[str, Any], client: Client) -> TextPost:
    return TextPost(d, client)

def load_link_post(d: Mapping[str, Any], client: Client) -> LinkPost:
    return LinkPost(d, client)

def load_gallery_post(d: Mapping[str, Any], client: Client) -> GalleryPost:
    return GalleryPost(d, client)

def load_poll_post(d: Mapping[str, Any], client: Client) -> PollPost:
    return PollPost(d, client)

def load_crosspost_submission(d: Mapping[str, Any], client: Client) -> CrosspostSubmission:
    return CrosspostSubmission(d, client)




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from .requisition import Requisition

from dataclasses import dataclass


@dataclass(repr=False, eq=False)
class SendParams:
    requisition: Requisition
    timeout: float = -2
    follow_redirects: Optional[bool] = None




from __future__ import annotations
from typing import TYPE_CHECKING, IO, Optional, Sequence, Any
if TYPE_CHECKING:
    from ...client_ASYNC import Client
    from ...models.widget.ASYNC import (
        TextAreaWidget,
        ButtonWidget,
        ImageWidget,
        CommunityListWidget,
        CalendarWidget,
        PostFlairWidget,
        CustomCSSWidget,
        CommunityDetailsWidget,
        ModeratorListWidget,
        RulesWidget,
        MenuBar,
    )
    from ...dtos.widget.button import (
        HoverState,
        TextHoverState,
        ImageHoverState,
        Button,
        TextButton,
        ImageButton,
    )
    from ...dtos.widget.custom_css import ImageInfo as CustomCSSWidgetImageInfo
    from ...dtos.widget.image import ImageInfo as ImageWidgetImageInfo
    from ...dtos.widget.menu_bar import (
        Tab,
        LinkTab,
        SubmenuTab,
    )
    from ...types import JSON_ro

import os.path as op
from functools import cached_property

from ...http.util.guess_filename_mimetype import guess_filename_mimetype
from ...models.widget.widget_image_upload_lease import WidgetImageUploadLease
from ...models.widget.ASYNC import WidgetList
from ...model_loaders.widget import load_widget_image_upload_lease
from ...model_loaders.widget_ASYNC import (
    load_widget,
    load_text_area_widget,
    load_button_widget,
    load_image_widget,
    load_community_list_widget,
    load_calendar_widget,
    load_post_flair_widget,
    load_custom_css_widget,
    load_community_details_widget,
    load_moderator_list_widget,
    load_rules_widget,
    load_menu_bar,
)


def _build_text_area_widget_json(
    *,
    header_color: str,
    background_color: str,
    title: str,
    text: str,
) -> JSON_ro:
    return {
        'kind': 'textarea',
        'styles': {'headerColor': header_color, 'backgroundColor': background_color},
        'shortName': title,
        'text': text,
    }

def _build_button_widget_json(
    *,
    header_color: str,
    background_color: str,
    title: str,
    description: str,
    buttons: Sequence[Button],
) -> JSON_ro:
    buttons_json: list[dict[str, JSON_ro]] = []
    for button in buttons:
        button_json: Optional[dict[str, JSON_ro]] = None
        if isinstance(button, TextButton):
            button_json = {
                'kind': 'text',
                'text': button.label,
                'url': button.link,
                'textColor': button.text_color,
                'fillColor': button.fill_color,
                'color': button.stroke_color,
            }
        elif isinstance(button, ImageButton):
            button_json = {
                'kind': 'image',
                'text': button.label,
                'linkUrl': button.link,
                'url': button.image_url,
                'width': button.image_size[0],
                'height': button.image_size[1],
            }
        else:
            if type(button) is Button:
                raise ValueError('a subclass of Button must be provided')
            raise Exception

        hs = button.hover_state
        if hs is not None:
            if isinstance(hs, TextHoverState):
                button_json['hoverState'] = {
                    'kind': 'text',
                    'text': hs.label,
                    'textColor': hs.text_color,
                    'fillColor': hs.fill_color,
                    'color': hs.stroke_color,
                }
            elif isinstance(hs, ImageHoverState):
                button_json['hoverState'] = {
                    'kind': 'image',
                    'url': hs.image_url,
                    'width': hs.image_size[0],
                    'height': hs.image_size[1],
                }
            else:
                if type(hs) is HoverState:
                    raise ValueError('a subclass of HoverState must be provided')
                raise Exception

        buttons_json.append(button_json)

    return {
        'kind': 'button',
        'styles': {'headerColor': header_color, 'backgroundColor': background_color},
        'shortName': title,
        'description': description,
        'buttons': buttons_json,
    }

def _build_image_widget_json(
    *,
    header_color: str,
    background_color: str,
    title: str,
    items: Sequence[ImageWidgetImageInfo],
) -> JSON_ro:
    return {
        'kind': 'image',
        'styles': {'headerColor': header_color, 'backgroundColor': background_color},
        'shortName': title,
        'data': [
            {
                'url': image_info.url,
                'width': image_info.size[0],
                'height': image_info.size[1],
                'linkUrl': image_info.link,
            }
            for image_info in items
        ],
    }

def _build_community_list_widget_json(
    *,
    header_color: str,
    background_color: str,
    title: str,
    items: Sequence[str],
) -> JSON_ro:
    return {
        'kind': 'community-list',
        'styles': {'headerColor': header_color, 'backgroundColor': background_color},
        'shortName': title,
        'data': list(items),
    }

def _build_calendar_widget_json(
    *,
    header_color: str,
    background_color: str,
    title: str,
    google_calendar_id: str,
    requires_sync: bool,
    num_events: int,
    show_title: bool,
    show_description: bool,
    show_location: bool,
    show_date: bool,
    show_time: bool,
) -> JSON_ro:
    return {
        'kind': 'calendar',
        'styles': {'headerColor': header_color, 'backgroundColor': background_color},
        'shortName': title,
        'googleCalendarId': google_calendar_id,
        'requiresSync': requires_sync,
        'configuration': {
            'numEvents': num_events,
            'showTitle': show_title,
            'showDescription': show_description,
            'showLocation': show_location,
            'showDate': show_date,
            'showTime': show_time,
        },
    }

def _build_post_flair_widget_json(
    *,
    header_color: str,
    background_color: str,
    title: str,
    display: str,
    order: Sequence[str],
) -> JSON_ro:
    return {
        'kind': 'post-flair',
        'styles': {'headerColor': header_color, 'backgroundColor': background_color},
        'shortName': title,
        'display': display,
        'order': list(order),
    }

def _build_custom_css_widget_json(
    *,
    header_color: str,
    background_color: str,
    title: str,
    text: str,
    css: str,
    height: Optional[int],
    image_data: Sequence[CustomCSSWidgetImageInfo],
) -> JSON_ro:
    return {
        'kind': 'custom',
        'styles': {'headerColor': header_color, 'backgroundColor': background_color},
        'shortName': title,
        'text': text,
        'css': css,
        'height': height,
        'imageData': [
            {
                'url': image_info.url,
                'width': image_info.width,
                'height': image_info.height,
                'name': image_info.name,
            }
            for image_info in image_data
        ],
    }

def _build_menu_bar_json(
    *,
    show_wiki: bool,
    tabs: Sequence[Tab],
) -> JSON_ro:
    tabs_json: list[dict[str, JSON_ro]] = []
    for tab in tabs:
        tab_json: Optional[dict[str, JSON_ro]] = None
        if isinstance(tab, LinkTab):
            tab_json = {'text': tab.label, 'url': tab.link}
        elif isinstance(tab, SubmenuTab):
            tab_json = {
                'text': tab.label,
                'children': [
                    {'text': smi.label, 'url': smi.link}
                    for smi in tab.items
                ],
            }
        else:
            if type(tab) is Tab:
                raise ValueError('a subclass of Tab must be provided')
            raise Exception

        tabs_json.append(tab_json)

    return {
        'kind': 'menu',
        'showWiki': show_wiki,
        'data': tabs_json,
    }


class WidgetProcedures:
    def __init__(self, client: Client) -> None:
        self._client = client

    class ImageUploading:
        def __init__(self, outer: WidgetProcedures) -> None:
            self._client = outer._client

        async def __call__(self,
            file: IO[bytes],
            *,
            sr: str,
            filepath: Optional[str] = None,
            timeout: float = 1000,
        ) -> WidgetImageUploadLease:
            return await self.upload(file, sr=sr, filepath=filepath, timeout=timeout)

        async def obtain_upload_lease(self,
            *,
            sr: str,
            filepath: str,
            mimetype: Optional[str] = None,
        ) -> WidgetImageUploadLease:
            if mimetype is None:
                mimetype = guess_filename_mimetype(filepath)
            result = await self._client.request('POST', f'/api/v1/{sr}/emoji_asset_upload_s3',
                    data={'filepath': filepath, 'mimetype': mimetype})
            return load_widget_image_upload_lease(result)

        async def deposit_file(self,
            file: IO[bytes],
            upload_lease: WidgetImageUploadLease,
            *,
            timeout: float = 1000,
        ) -> None:
            resp = await self._client.http.request('POST', upload_lease.endpoint,
                    data=upload_lease.fields, files={'file': file}, timeout=timeout)
            resp.ensure_successful_status()

        async def upload(self,
            file: IO[bytes],
            *,
            sr: str,
            filepath: Optional[str] = None,
            timeout: float = 1000,
        ) -> WidgetImageUploadLease:
            if filepath is None:
                filepath = op.basename(getattr(file, 'name', ''))
                if not filepath:
                    raise ValueError("the `filepath` parameter must be explicitly specified if the file object has no `name` attribute.")
            upload_lease = await self.obtain_upload_lease(sr=sr, filepath=filepath)
            await self.deposit_file(file, upload_lease, timeout=timeout)
            return upload_lease

    image_uploading: cached_property[ImageUploading] = cached_property(ImageUploading)

    class Create:
        def __init__(self, outer: WidgetProcedures) -> None:
            self._client = outer._client

        async def _invoke(self, sr: str, json: JSON_ro) -> Any:
            return await self._client.request('POST', f'/r/{sr}/api/widget', json=json)

        async def text_area(self,
            sr: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            text: str,
        ) -> TextAreaWidget:
            json = _build_text_area_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                text=text,
            )
            root = await self._invoke(sr, json=json)
            return load_text_area_widget(root)

        async def button(self,
            sr: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            description: str,
            buttons: Sequence[Button],
        ) -> ButtonWidget:
            json = _build_button_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                description=description,
                buttons=buttons,
            )
            root = await self._invoke(sr, json=json)
            return load_button_widget(root)

        async def image(self,
            sr: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            items: Sequence[ImageWidgetImageInfo],
        ) -> ImageWidget:
            json = _build_image_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                items=items,
            )
            root = await self._invoke(sr, json=json)
            return load_image_widget(root)

        async def community_list(self,
            sr: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            items: Sequence[str],
        ) -> CommunityListWidget:
            json = _build_community_list_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                items=items,
            )
            root = await self._invoke(sr, json=json)
            return load_community_list_widget(root)

        async def calendar(self,
            sr: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            google_calendar_id: str,
            requires_sync: bool,
            num_events: int,
            show_title: bool,
            show_description: bool,
            show_location: bool,
            show_date: bool,
            show_time: bool,
        ) -> CalendarWidget:
            json = _build_calendar_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                google_calendar_id=google_calendar_id,
                requires_sync=requires_sync,
                num_events=num_events,
                show_title=show_title,
                show_description=show_description,
                show_location=show_location,
                show_date=show_date,
                show_time=show_time,
            )
            root = await self._invoke(sr, json=json)
            return load_calendar_widget(root)

        async def post_flair(self,
            sr: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            display: str,
            order: Sequence[str],
        ) -> PostFlairWidget:
            json = _build_post_flair_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                display=display,
                order=order,
            )
            root = await self._invoke(sr, json=json)
            return load_post_flair_widget(root)

        async def custom_css(self,
            sr: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            text: str,
            css: str,
            height: Optional[int],
            image_data: Sequence[CustomCSSWidgetImageInfo],
        ) -> CustomCSSWidget:
            json = _build_custom_css_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                text=text,
                css=css,
                height=height,
                image_data=image_data,
            )
            root = await self._invoke(sr, json=json)
            return load_custom_css_widget(root)

    create: cached_property[Create] = cached_property(Create)

    class Replace:
        def __init__(self, outer: WidgetProcedures) -> None:
            self._client = outer._client

        async def _invoke(self, sr: str, idt: str, json: JSON_ro) -> Any:
            return await self._client.request('PUT', f'/r/{sr}/api/widget/{idt}', json=json)

        async def text_area(self,
            sr: str,
            idt: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            text: str,
        ) -> TextAreaWidget:
            json = _build_text_area_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                text=text,
            )
            root = await self._invoke(sr, idt, json=json)
            return load_text_area_widget(root)

        async def button(self,
            sr: str,
            idt: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            description: str,
            buttons: Sequence[Button],
        ) -> ButtonWidget:
            json = _build_button_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                description=description,
                buttons=buttons,
            )
            root = await self._invoke(sr, idt, json=json)
            return load_button_widget(root)

        async def image(self,
            sr: str,
            idt: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            items: Sequence[ImageWidgetImageInfo],
        ) -> ImageWidget:
            json = _build_image_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                items=items,
            )
            root = await self._invoke(sr, idt, json=json)
            return load_image_widget(root)

        async def community_list(self,
            sr: str,
            idt: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            items: Sequence[str],
        ) -> CommunityListWidget:
            json = _build_community_list_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                items=items,
            )
            root = await self._invoke(sr, idt, json=json)
            return load_community_list_widget(root)

        async def calendar(self,
            sr: str,
            idt: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            google_calendar_id: str,
            requires_sync: bool,
            num_events: int,
            show_title: bool,
            show_description: bool,
            show_location: bool,
            show_date: bool,
            show_time: bool,
        ) -> CalendarWidget:
            json = _build_calendar_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                google_calendar_id=google_calendar_id,
                requires_sync=requires_sync,
                num_events=num_events,
                show_title=show_title,
                show_description=show_description,
                show_location=show_location,
                show_date=show_date,
                show_time=show_time,
            )
            root = await self._invoke(sr, idt, json=json)
            return load_calendar_widget(root)

        async def post_flair(self,
            sr: str,
            idt: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            display: str,
            order: Sequence[str],
        ) -> PostFlairWidget:
            json = _build_post_flair_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                display=display,
                order=order,
            )
            root = await self._invoke(sr, idt, json=json)
            return load_post_flair_widget(root)

        async def custom_css(self,
            sr: str,
            idt: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            text: str,
            css: str,
            height: Optional[int],
            image_data: Sequence[CustomCSSWidgetImageInfo],
        ) -> CustomCSSWidget:
            json = _build_custom_css_widget_json(
                header_color=header_color,
                background_color=background_color,
                title=title,
                text=text,
                css=css,
                height=height,
                image_data=image_data,
            )
            root = await self._invoke(sr, idt, json=json)
            return load_custom_css_widget(root)

        async def community_details(self,
            sr: str,
            idt: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            subscriber_text: str,
            viewing_text: str,
        ) -> CommunityDetailsWidget:
            json: JSON_ro = {
                'kind': 'post-flair',
                'styles': {'headerColor': header_color, 'backgroundColor': background_color},
                'shortName': title,
                'subscribersText': subscriber_text,
                'currentlyViewingText': viewing_text,
            }
            root = await self._invoke(sr, idt, json=json)
            return load_community_details_widget(root)

        async def moderator_list(self,
            sr: str,
            idt: str,
            *,
            header_color: str,
            background_color: str,
        ) -> ModeratorListWidget:
            json: JSON_ro = {
                'kind': 'moderators',
                'styles': {'headerColor': header_color, 'backgroundColor': background_color},
            }
            root = await self._invoke(sr, idt, json=json)
            return load_moderator_list_widget(root)

        async def rules(self,
            sr: str,
            idt: str,
            *,
            header_color: str,
            background_color: str,
            title: str,
            display: str,
        ) -> RulesWidget:
            json: JSON_ro = {
                'kind': 'post-flair',
                'styles': {'headerColor': header_color, 'backgroundColor': background_color},
                'shortName': title,
                'display': display,
            }
            root = await self._invoke(sr, idt, json=json)
            return load_rules_widget(root)

    replace: cached_property[Replace] = cached_property(Replace)

    async def create_menu_bar_tabs(self,
        sr: str,
        *,
        show_wiki: bool,
        tabs: Sequence[Tab],
    ) -> MenuBar:
        json = _build_menu_bar_json(
            show_wiki=show_wiki,
            tabs=tabs,
        )
        root = await self._client.request('POST', f'/r/{sr}/api/widget', json=json)
        return load_menu_bar(root)

    async def replace_menu_bar_tabs(self,
        sr: str,
        idt: str,
        *,
        show_wiki: bool,
        tabs: Sequence[Tab],
    ) -> MenuBar:
        json = _build_menu_bar_json(
            show_wiki=show_wiki,
            tabs=tabs,
        )
        root = await self._client.request('PUT', f'/r/{sr}/api/widget/{idt}', json=json)
        return load_menu_bar(root)

    async def retrieve(self, sr: str) -> WidgetList:
        root = await self._client.request('GET', f'/r/{sr}/api/widgets')
        layout = root['layout']

        order = layout['sidebar']['order']
        object_map = root['items']

        widgets = [load_widget(object_map[idt]) for idt in order]

        menu_bar = None
        topbar_order = layout['topbar']['order']
        if topbar_order:
            menu_bar = load_menu_bar(object_map[topbar_order[0]])

        id_card_widget_id = layout['idCardWidget']
        community_details_widget = load_community_details_widget(object_map[id_card_widget_id])

        moderator_list_widget = load_moderator_list_widget(object_map[layout['moderatorWidget']])

        subr_id36 = id_card_widget_id.rpartition('-')[2]
        rules_widget = None
        if (d := object_map.get(f'widget_rules-{subr_id36}')) is not None:
            rules_widget = load_rules_widget(d)

        return WidgetList(
            widgets=widgets,
            menu_bar=menu_bar,
            community_details_widget=community_details_widget,
            moderator_list_widget=moderator_list_widget,
            rules_widget=rules_widget,
        )

    async def delete(self, sr: str, idt: str) -> None:
        await self._client.request('DELETE', f'/r/{sr}/api/widget/{idt}')

    async def reorder(self, sr: str, order: Sequence[str]) -> None:
        await self._client.request('PATCH', f'/r/{sr}/api/widget_order/sidebar', json=list(order))




from __future__ import annotations
from typing import TYPE_CHECKING, Mapping, Any, Optional, MutableSequence
if TYPE_CHECKING:
    from ..client_SYNC import Client

from ..models.modmail_SYNC import (
    ConversationInfo,
    Message,
    ModAction,
    UserDossier,
    ConversationAggregate,
    UserDossierConversationAggregate,
    OptionalUserDossierConversationAggregate,
)

def load_conversation_info(d: Mapping[str, Any], client: Client) -> ConversationInfo:
    return ConversationInfo(d, client)

def load_message(d: Mapping[str, Any], client: Client) -> Message:
    return Message(d, client)

def load_mod_action(d: Mapping[str, Any]) -> ModAction:
    return ModAction(d)

def load_user_dossier(d: Mapping[str, Any]) -> UserDossier:
    return UserDossier(d)


def load_conversation_aggregate(
    conversation_data: Mapping[str, Any],
    messages_mapping_data: Mapping[str, Any],
    mod_actions_mapping_data: Mapping[str, Any],
    *,
    client: Client,
) -> ConversationAggregate:
    conversation = load_conversation_info(conversation_data, client)

    messages: MutableSequence[Message] = []
    mod_actions: MutableSequence[ModAction] = []
    history: MutableSequence[object] = []

    convo_references = conversation_data['objIds']
    for convo_ref in convo_references:
        key = convo_ref['key']
        id36 = convo_ref['id']

        convo_obj: object = None
        if key == 'messages':
            message_data = messages_mapping_data[id36]
            convo_obj = load_message(message_data, client)
            messages.append(convo_obj)
        elif key == 'modActions':
            mod_action_data = mod_actions_mapping_data[id36]
            convo_obj = load_mod_action(mod_action_data)
            mod_actions.append(convo_obj)

        if convo_obj is None:
            raise Exception('unknown modmail object type encountered')
        history.append(convo_obj)

    return ConversationAggregate(conversation, messages, mod_actions, history)

def load_user_dossier_conversation_aggregate(
    conversation_data: Mapping[str, Any],
    messages_mapping_data: Mapping[str, Any],
    mod_actions_mapping_data: Mapping[str, Any],
    user_dossier_data: Mapping[str, Any],
    *,
    client: Client,
) -> UserDossierConversationAggregate:
    conversation = load_conversation_info(conversation_data, client)
    user_dossier = load_user_dossier(user_dossier_data)

    messages: MutableSequence[Message] = []
    mod_actions: MutableSequence[ModAction] = []
    history: MutableSequence[object] = []

    convo_references = conversation_data['objIds']
    for convo_ref in convo_references:
        key = convo_ref['key']
        id36 = convo_ref['id']

        convo_obj: object = None
        if key == 'messages':
            message_data = messages_mapping_data[id36]
            convo_obj = load_message(message_data, client)
            messages.append(convo_obj)
        elif key == 'modActions':
            mod_action_data = mod_actions_mapping_data[id36]
            convo_obj = load_mod_action(mod_action_data)
            mod_actions.append(convo_obj)

        if convo_obj is None:
            raise Exception('unknown modmail object type encountered')
        history.append(convo_obj)

    return UserDossierConversationAggregate(conversation, messages, mod_actions, history, user_dossier)

def load_optional_user_dossier_conversation_aggregate(
    conversation_data: Mapping[str, Any],
    messages_mapping_data: Mapping[str, Any],
    mod_actions_mapping_data: Mapping[str, Any],
    optional_user_dossier_data: Optional[Mapping[str, Any]],
    *,
    client: Client,
) -> OptionalUserDossierConversationAggregate:
    conversation = load_conversation_info(conversation_data, client)
    user_dossier = None
    if optional_user_dossier_data is not None:
        user_dossier = load_user_dossier(optional_user_dossier_data)

    messages: MutableSequence[Message] = []
    mod_actions: MutableSequence[ModAction] = []
    history: MutableSequence[object] = []

    convo_references = conversation_data['objIds']
    for convo_ref in convo_references:
        key = convo_ref['key']
        id36 = convo_ref['id']

        convo_obj: object = None
        if key == 'messages':
            message_data = messages_mapping_data[id36]
            convo_obj = load_message(message_data, client)
            messages.append(convo_obj)
        elif key == 'modActions':
            mod_action_data = mod_actions_mapping_data[id36]
            convo_obj = load_mod_action(mod_action_data)
            mod_actions.append(convo_obj)

        if convo_obj is None:
            raise Exception('unknown modmail object type encountered')
        history.append(convo_obj)

    return OptionalUserDossierConversationAggregate(conversation, messages, mod_actions, history, user_dossier)




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Mapping
if TYPE_CHECKING:
    from ..client_SYNC import Client
    from ..models.more_comments_SYNC import MoreComments
    from ..models.comment_SYNC import Comment

from ..models.comment_tree_SYNC import MoreCommentsTreeNode, CommentTreeNode, SubmissionTreeNode
from ..models.more_comments_SYNC import ContinueThisThread, LoadMoreComments
from .comment_SYNC import load_comment
from .submission_SYNC import load_submission

def load_more_comments(
    d: Mapping[str, Any],
    client: Client,
    submission_id36: str,
    sort: str,
) -> MoreComments:
    _, _, comment_id36 = d['parent_id'].partition('_')
    go_deeper = d['id'] == '_'
    if go_deeper:
        return ContinueThisThread(
            submission_id36=submission_id36,
            comment_id36=comment_id36,
            sort=sort,
            d=d,
            client=client,
        )
    return LoadMoreComments(
        submission_id36=submission_id36,
        comment_id36=comment_id36,
        child_id36s=d['children'],
        sort=sort,
        count=d['count'],
        d=d,
        client=client,
    )


def load_submission_tree_node(d: Any, client: Client, sort: str) -> SubmissionTreeNode:
    def load_comment_tree_node(d: Any, client: Client, submission_id36: str, sort: str) -> CommentTreeNode:
        value = load_comment(d, client)
        children = []
        more = None

        replies_data = d['replies']
        if replies_data:
            elements = list(replies_data['data']['children'])
            if elements:
                m = elements[-1]
                if m['kind'] == 'more':
                    elements.pop()
                    more = load_more_comments(m['data'], client, submission_id36, sort)

                for m in elements:
                    node = load_comment_tree_node(m['data'], client, submission_id36, sort)
                    children.append(node)

        return CommentTreeNode(value, children, more)

    submission_data = d[0]['data']['children'][0]['data']
    submission_id36: str = submission_data['id']

    value = load_submission(submission_data, client)
    children = []
    more = None

    elements = list(d[1]['data']['children'])
    if elements:
        m = elements[-1]
        if m['kind'] == 'more':
            elements.pop()
            more = load_more_comments(m['data'], client, submission_id36, sort)

        for m in elements:
            node = load_comment_tree_node(m['data'], client, submission_id36, sort)
            children.append(node)

    return SubmissionTreeNode(value, children, more)


def load_more_comments_tree_node(d: Any, client: Client, submission_id36: str, sort: str) -> MoreCommentsTreeNode:
    value_lookup: dict[str, Comment] = {}
    children_lookup: dict[str, list[CommentTreeNode]] = {}
    more_lookup: dict[str, MoreComments] = {}

    elements = d['json']['data']['things']
    for m in elements:
        data = m['data']
        kind = m['kind']
        if kind == 'more':
            more = load_more_comments(data, client, submission_id36, sort)
            parent_idt = data['parent_id']
            more_lookup[parent_idt] = more
        else:
            value = load_comment(data, client)
            children: list[CommentTreeNode] = []
            idt = data['name']
            value_lookup[idt] = value
            children_lookup[idt] = children

    root_children: list[CommentTreeNode] = []
    for m in elements:
        data = m['data']
        kind = m['kind']
        if kind != 'more':
            idt = data['name']
            node = CommentTreeNode(
                value_lookup[idt],
                children_lookup[idt],
                more_lookup.pop(idt, None),
            )
            parent_idt = data['parent_id']
            children_lookup.get(parent_idt, root_children).append(node)

    root_more = None
    if more_lookup:
        root_more = more_lookup.popitem()[1]
    if more_lookup:
        raise Exception

    return MoreCommentsTreeNode(None, root_children, root_more)



"""HTTP transport connectors.

Each of the modules in this subpackage represents an HTTP library. Transport connectors
are HTTP libraries that can be used as drivers to make HTTP requests with this library.

When you import a module from this subpackage, the transport library is loaded and
registered as an available transport connector. You can explicitly import a transport
connector module in your program's `__main__` module to have RedditWarp prioritise that
transport connector. If you do this and the transport package is not installed, an
`ImportError` will occur.

Each transport connector module defines a connector class and a `new_connector` function
to construct a connector object from zero arguments. The transport's name is in `name`
and its version string is available in `version`.
"""




from __future__ import annotations
from typing import TYPE_CHECKING, Optional, Callable, MutableMapping
if TYPE_CHECKING:
    from ..auth.token_obtainment_client_ASYNC import TokenObtainmentClient
    from ..auth.token import Token
    from ..http.handler_ASYNC import Handler
    from ..http.requisition import Requisition
    from ..http.send_params import SendParams
    from ..http.exchange import Exchange

from ..util.imports import lazy_import
if TYPE_CHECKING:
    import asyncio
else:
    lazy_import % 'asyncio'

import time

from ..http.delegating_handler_ASYNC import DelegatingHandler
from ..auth.exceptions import (
    UnknownTokenType,
    extract_www_authenticate_auth_params,
    raise_for_resource_server_response_error,
)


class Authorizer:
    """An object for managing the token.

    The `Authorizer` keeps token information and knows how to renew the token
    when it expires and how to prepare a requisition.
    """

    def __init__(self,
        token_client: Optional[TokenObtainmentClient] = None,
        token: Optional[Token] = None,
    ) -> None:
        self.token_client: Optional[TokenObtainmentClient] = token_client
        ("""
            A token obtainment client used to make the token obtainment requests.
            """)
        self.token: Optional[Token] = token
        ("""
            Holds the current token information.
            """)
        self.renewal_time: Optional[float] = None
        ("""
            Time at which the token is expected to expire. See :meth:`.time`.
            """)
        self.renewal_skew: float = 30
        ("""
            Number of seconds before the token expiration when the token should be renewed.
            """)
        self.expires_in_fallback: Optional[int] = None
        ("""
            A fallback value for the token expiration time when not specified by the server.
            """)
        self.time_func: Callable[[], float] = time.monotonic
        ("""
            A function that returns the current time. The :meth:`time` method calls this this function.
            """)
        self.authorization_header_name: str = 'Authorization'
        ("""
            Name of the `Authorization` header. Use this attribute to change its capitalization.
            """)

    def has_token_client(self) -> bool:
        return self.token_client is not None

    def fetch_token_client(self) -> TokenObtainmentClient:
        """Return `.token_client`. Raise `RuntimeError` if not set."""
        v = self.token_client
        if v is None:
            raise RuntimeError('token client not set')
        return v

    def set_token_client(self, value: Optional[TokenObtainmentClient]) -> None:
        """Set `.token_client`."""
        self.token_client = value

    def has_token(self) -> bool:
        return self.token is not None

    def fetch_token(self) -> Token:
        """Return `.token`. Raise `RuntimeError` if not set."""
        v = self.token
        if v is None:
            raise RuntimeError('token not set')
        return v

    def set_token(self, value: Optional[Token]) -> None:
        """Set `.token`."""
        self.token = value

    async def renew_token(self) -> None:
        """Renew the token.

        .. .RAISES

        :raises RuntimeError:
            No token client (:attr:`token_client`) is set.
        """
        tc = self.fetch_token_client()
        tk = await tc.fetch_token()
        if tk.token_type.lower() != 'bearer':
            raise UnknownTokenType(token=tk)

        self.set_token(tk)

        expires_in = tk.expires_in
        if expires_in is None:
            expires_in = self.expires_in_fallback
        if expires_in is None:
            self.renewal_time = None
        else:
            self.renewal_time = self.time() + expires_in - self.renewal_skew

        if tk.refresh_token:
            grant1 = tc.grant
            if (
                grant1.get('grant_type', '') == 'refresh_token'
                and grant1.get('refresh_token', '') != tk.refresh_token
            ):
                tc.grant = {**grant1, 'refresh_token': tk.refresh_token}

    def time(self) -> float:
        """Return the current internal time. This is used for :attr:`renewal_time`."""
        return self.time_func()

    def should_renew_token(self) -> bool:
        """Return true if the token is expired, about to expire, or is unset."""
        if not self.has_token():
            return True
        if self.renewal_time is None:
            return False
        return self.time() >= self.renewal_time

    def can_renew_token(self) -> bool:
        """The token can be renewed if a token client (:attr:`token_client`) is available."""
        return self.has_token_client()

    async def attain_token(self) -> Token:
        if self.should_renew_token():
            await self.renew_token()
        return self.fetch_token()

    def get_header_entry(self) -> tuple[str, str]:
        """Return an authorization header entry tuple."""
        return (self.authorization_header_name, '{0.token_type} {0.access_token}'.format(self.fetch_token()))

    def prepare_requisition(self, requisition: Requisition) -> None:
        """Prepare a requisition for authorization using the token."""
        self.prepare_headers(requisition.headers)

    def prepare_headers(self, headers: MutableMapping[str, str]) -> None:
        """Update headers with an authorization entry."""
        k, v = self.get_header_entry()
        headers[k] = v


class Authorized(DelegatingHandler):
    """Used to make requests to endpoints that require authorization."""

    def __init__(self, handler: Handler, authorizer: Authorizer) -> None:
        super().__init__(handler)
        self.authorizer: Authorizer = authorizer
        ("")
        self._lock = asyncio.Lock()

    async def _send(self, p: SendParams) -> Exchange:
        authorizer = self.authorizer
        async with self._lock:
            if authorizer.should_renew_token():
                await authorizer.renew_token()

        used_token = authorizer.fetch_token()
        authorizer.prepare_requisition(p.requisition)
        xchg = await super()._send(p)
        auth_params = extract_www_authenticate_auth_params(xchg.response)

        invalid_token = auth_params.get('error', '') == 'invalid_token'
        if invalid_token and authorizer.can_renew_token():
            async with self._lock:
                if used_token.access_token == authorizer.fetch_token().access_token:
                    await authorizer.renew_token()

            authorizer.prepare_requisition(p.requisition)
            xchg = await super()._send(p)
            auth_params = extract_www_authenticate_auth_params(xchg.response)

        raise_for_resource_server_response_error(auth_params)

        return xchg



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['redditwarp',
 'redditwarp._cli',
 'redditwarp.auth',
 'redditwarp.core',
 'redditwarp.dark',
 'redditwarp.dark.auth',
 'redditwarp.dark.core',
 'redditwarp.dark.siteprocs',
 'redditwarp.dark.siteprocs.login',
 'redditwarp.dark.util',
 'redditwarp.dtos',
 'redditwarp.dtos.widget',
 'redditwarp.http',
 'redditwarp.http.misc',
 'redditwarp.http.transport',
 'redditwarp.http.transport.connectors',
 'redditwarp.http.util',
 'redditwarp.iterators',
 'redditwarp.model_loaders',
 'redditwarp.models',
 'redditwarp.models.widget',
 'redditwarp.pagination',
 'redditwarp.pagination.paginators',
 'redditwarp.pagination.paginators.listing',
 'redditwarp.pagination.paginators.listing.mixins',
 'redditwarp.pagination.paginators.moderation',
 'redditwarp.pagination.paginators.user',
 'redditwarp.pushshift',
 'redditwarp.pushshift.core',
 'redditwarp.pushshift.models',
 'redditwarp.pushshift.paginators',
 'redditwarp.pushshift.siteprocs',
 'redditwarp.siteprocs',
 'redditwarp.siteprocs.account',
 'redditwarp.siteprocs.collection',
 'redditwarp.siteprocs.comment',
 'redditwarp.siteprocs.comment_tree',
 'redditwarp.siteprocs.custom_feed',
 'redditwarp.siteprocs.draft',
 'redditwarp.siteprocs.flair',
 'redditwarp.siteprocs.flair_emoji',
 'redditwarp.siteprocs.front',
 'redditwarp.siteprocs.live_thread',
 'redditwarp.siteprocs.message',
 'redditwarp.siteprocs.misc',
 'redditwarp.siteprocs.moderation',
 'redditwarp.siteprocs.modmail',
 'redditwarp.siteprocs.submission',
 'redditwarp.siteprocs.subreddit',
 'redditwarp.siteprocs.subreddit_style_new',
 'redditwarp.siteprocs.subreddit_style_old',
 'redditwarp.siteprocs.user',
 'redditwarp.siteprocs.widget',
 'redditwarp.siteprocs.wiki',
 'redditwarp.streaming',
 'redditwarp.streaming.makers',
 'redditwarp.util',
 'redditwarp.websocket',
 'redditwarp.websocket.transport',
 'redditwarp.websocket.transport.carriers']

package_data = \
{'': ['*']}

setup_kwargs = {
    'name': 'redditwarp',
    'version': '1.0.0rc0',
    'description': 'A library for interacting with the Reddit API.',
    'long_description': '\n# RedditWarp\n\n[![Latest version](https://img.shields.io/pypi/v/redditwarp)](https://pypi.org/p/redditwarp/)\n![Supported Python versions](https://img.shields.io/pypi/pyversions/redditwarp)\n[![Discord guild](https://img.shields.io/discord/1084793643673600062?logo=discord&logoColor=white&logoWidth=20&label=&labelColor=5865F2&color=7289DA)](http://pyprohly.github.io/redditwarp/discord-invite)\n\nA comprehensive, type-complete, easy-to-learn Python Reddit API wrapper.\n\nRedditWarp is a Python library that simplifies working with the Reddit API.\nIt handles the complexities of the Reddit API in a way that is comprehensive,\nhighly discoverable, and static-type conscious. It includes a variety of useful\ntools to facilitate a wide range of use cases, and even provides tools for\naccessing the API in ways previously not possible.\n\nLook how easy it is to use:\n\n```python\nimport redditwarp.SYNC\n\nclient = redditwarp.SYNC.Client()\n\nit = client.p.front.pull.hot(5)\nl = list(it)\nfor subm in l:\n    print("r/{0.subreddit.name} | {0.id36}+ ^{0.score} | {0.title!r:.80}".format(subm))\n```\n\n## Features\n\n* A consistent and easy-to-use programming interface.\n* Modern type-complete codebase.\n* Sync and asynchronous IO support.\n* Automatic rate limit handling.\n* A comprehensive and discoverable index of API procedures.\n* Model classes that sensibly wrap API structures.\n* Formal data structures to facilitate comment tree traversals and pagination.\n* HTTP transport library agnosticism.\n* OAuth2 tooling and CLI utilities to help manage auth tokens.\n* A simple event-based listing endpoint streaming framework.\n\n## Installation\n\n**Requires Python 3.8+.**\nType annotations may use 3.9 features.\nCode examples will assume 3.10.\n\nInstall/update:\n\n    pip install -U redditwarp\n\n## Demonstration\n\n<details open>\n  <summary>Examples</summary>\n\n```python\nimport redditwarp.SYNC\n\nclient = redditwarp.SYNC.Client()\n\n# Display the first 5 submissions on the Reddit frontpage.\nit = client.p.front.pull.hot(5)\nl = list(it)\nfor subm in l:\n    print("r/{0.subreddit.name} | {0.id36}+ ^{0.score} | {0.title!r:.80}".format(subm))\n\n# How many subscribers does r/Python have?\nsubr = client.p.subreddit.fetch_by_name(\'Python\')\nprint(subr.subscriber_count)\n\n# Display the top submission of the week in the r/YouShouldKnow subreddit.\nm = next(client.p.subreddit.pull.top(\'YouShouldKnow\', amount=1, time=\'week\'))\nprint(f\'\'\'\\\n{m.permalink}\n{m.id36}+ ^{m.score} | {m.title}\nSubmitted {m.created_at.astimezone().ctime()}{\' *\' if m.is_edited else \'\'} \\\nby u/{m.author_display_name} to r/{m.subreddit.name}\n\'\'\')\n\n# Get the first comment of a submission.\ntree_node = client.p.comment_tree.fetch(int(\'uc8i1g\', 36), sort=\'top\', limit=1)\nc = tree_node.children[0].value\nprint(f\'\'\'\\\n{c.submission.id36}+{c.id36} ^{c.score}\nu/{c.author_display_name} says:\n{c.body}\n\'\'\')\n\n# List the moderators of r/redditdev.\nit1 = client.p.moderation.pull_users.moderators(\'redditdev\')\nfor mod in it1:\n    print(mod.name)\n```\n\n</details>\n\n<details>\n  <summary>More examples</summary>\n\n```python\n# ...\n\n# Need credentials for these next few API calls.\nCLIENT_ID = \'...\'\nCLIENT_SECRET = \'...\'\nREFRESH_TOKEN = \'...\'\nclient1 = redditwarp.SYNC.Client(CLIENT_ID, CLIENT_SECRET, REFRESH_TOKEN)\n\n# Who am I?\nme = client1.p.account.fetch()\nprint(f"Hello u/{me.name}!")\n\n# Show my last 5 comments.\nit2 = client.p.user.pull.comments(me.name, 5)\nfor comm in it2:\n    print(\'###\')\n    print(comm.body)\n\n# Show my last 10 saved items.\nfrom redditwarp.models.submission_SYNC import Submission\nfrom redditwarp.models.comment_SYNC import Comment\nit3 = client1.p.user.pull.saved(me.name, 10)\nl = list(it3)\nfor obj in l:\n    print(\'###\')\n    match obj:\n        case Submission() as m:\n            print(f\'\'\'\\\n{m.permalink}\n{m.id36}+ ^{m.score} | {m.title}\nSubmitted {m.created_at.astimezone().ctime()}{\' *\' if m.is_edited else \'\'} \\\nby u/{m.author_display_name} to r/{m.subreddit.name}\n\'\'\')\n        case Comment() as c:\n            print(f\'\'\'\\\n{c.permalink}\n{c.submission.id36}+{c.id36} ^{c.score}\nu/{c.author_display_name} says:\n{c.body}\n\'\'\')\n\n# Submit a link post to r/test.\nsubm_id = client1.p.submission.create_link_post(\'test\',\n        "Check out this cool website", "https://www.reddit.com")\n\n# Reply to a submission.\nfrom redditwarp.util.extract_id_from_url import extract_submission_id_from_url\nidn = extract_submission_id_from_url("https://www.reddit.com/comments/5e1az9")\ncomm1 = client1.p.submission.reply(idn, "Pretty cool stuff!")\n\n# Delete the post and the comment reply.\nclient1.p.submission.delete(subm_id)\nclient1.p.comment.delete(comm1.id)\n```\n\n</details>\n\n<details>\n  <summary>Streaming example</summary>\n\n```python\n#!/usr/bin/env python\nfrom __future__ import annotations\nfrom typing import TYPE_CHECKING\nif TYPE_CHECKING:\n    from redditwarp.models.submission_ASYNC import Submission\n\nimport asyncio\n\nimport redditwarp.ASYNC\nfrom redditwarp.streaming.makers.subreddit_ASYNC import create_submission_stream\nfrom redditwarp.streaming.ASYNC import flow\n\nasync def main() -> None:\n    client = redditwarp.ASYNC.Client()\n    async with client:\n        submission_stream = create_submission_stream(client, \'AskReddit\')\n\n        @submission_stream.output.attach\n        async def _(subm: Submission) -> None:\n            print(subm.id36, \'~\', subm.title)\n\n        @submission_stream.error.attach\n        async def _(exc: Exception) -> None:\n            print(\'ERROR:\', repr(exc))\n\n        await flow(submission_stream)\n\nasyncio.run(main())\n```\n\n</details>\n\n## Support\n\nPost any questions you have to [r/redditdev].\n\n[r/redditdev]: https://www.reddit.com/r/redditdev/\n\nJoin the conversation in the RedditWarp [Discord guild].\n\n[Discord guild]: http://pyprohly.github.io/redditwarp/discord-invite\n\n## Links\n\n* [Repository](https://github.com/Pyprohly/redditwarp)\n* [PyPI](https://pypi.org/p/redditwarp/)\n* [Documentation](https://redditwarp.readthedocs.io/)\n* [r/redditdev]\n* [Discord guild]\n',
    'author': 'Pyprohly',
    'author_email': 'pyprohly@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



"""A general purpose event scheduler.

This is just a reimplementation of the `sched` module from the standard library,
but the names of the methods are different.
"""

from __future__ import annotations
from typing import Callable

from collections.abc import Sized
import time
import heapq


class Entry:
    def __init__(self, when: float, callback: Callable[[], None]) -> None:
        self.when: float = when
        ("")
        self.callback: Callable[[], None] = callback
        ("")
        self.scheduled: bool = True
        ("")
        self.cancelled: bool = False
        ("")

    def __eq__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.when == other.when
        return NotImplemented
    def __lt__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.when < other.when
        return NotImplemented
    def __le__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.when <= other.when
        return NotImplemented
    def __gt__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.when > other.when
        return NotImplemented
    def __ge__(self, other: object) -> bool:
        if isinstance(other, self.__class__):
            return self.when >= other.when
        return NotImplemented

class Handle:
    def __init__(self, entry: Entry) -> None:
        self._entry = entry

    def get_when(self) -> float:
        return self._entry.when
    def get_callback(self) -> Callable[[], None]:
        return self._entry.callback
    def is_scheduled(self) -> bool:
        return self._entry.scheduled
    def is_cancelled(self) -> bool:
        return self._entry.cancelled

    def cancel(self) -> None:
        self._entry.cancelled = True


class Scheduler(Sized):
    def __init__(self,
        *,
        time_func: Callable[[], float] = time.monotonic,
        delay_func: Callable[[float], None] = time.sleep,
    ) -> None:
        self.time_func: Callable[[], float] = time_func
        ("")
        self.delay_func: Callable[[float], None] = delay_func
        ("")
        self._pq: list[Entry] = []

    def __len__(self) -> int:
        return len(self._pq)

    def empty(self) -> bool:
        return not self._pq

    def time(self) -> float:
        return self.time_func()

    def call_at(self, when: float, callback: Callable[[], None]) -> Handle:
        entry = Entry(when, callback)
        heapq.heappush(self._pq, entry)
        return Handle(entry)

    def call_later(self, delay: float, callback: Callable[[], None]) -> Handle:
        return self.call_at(self.time() + delay, callback)

    def call_soon(self, callback: Callable[[], None]) -> Handle:
        return self.call_later(0, callback)

    def jog(self) -> float:
        """Run the scheduler while there are immediate events to be processed.

        Return the amount of time it will take until the next event is ready.
        If there are no further scheduled events, the value `-1` is returned.
        """
        pq = self._pq
        while pq:
            entry = pq[0]
            if entry.cancelled:
                heapq.heappop(pq)
                entry.scheduled = False
                continue

            curr_time = self.time_func()
            if entry.when > curr_time:
                return entry.when - curr_time

            heapq.heappop(pq)
            entry.scheduled = False
            entry.callback()
        return -1

    def run(self) -> None:
        """Run the scheduller until all scheduled events are complete."""
        while True:
            t = self.jog()
            if t < 0:
                break
            self.delay_func(t)




from __future__ import annotations
from typing import Iterator

from ..util.shdlr import Scheduler


def flow(*streams: Iterator[float]) -> None:
    flow_series(*streams)

def flow_series(*streams: Iterator[float]) -> None:
    def _invoke_and_reschedule(itr: Iterator[float], shdlr: Scheduler) -> None:
        try:
            t = next(itr)
        except StopIteration:
            return
        shdlr.call_later(t, (lambda: _invoke_and_reschedule(itr, shdlr)))

    shdlr = Scheduler()
    for itr in streams:
        _invoke_and_reschedule(itr, shdlr)
    shdlr.run()




from __future__ import annotations
from typing import TYPE_CHECKING, Optional
if TYPE_CHECKING:
    from ...client_SYNC import Client
    from ...models.comment_SYNC import LooseComment
    from ...models.submission_SYNC import Submission

from functools import cached_property

from ...pagination.paginator_chaining_iterator import ImpartedPaginatorChainingIterator
from ...pagination.paginators.user.sync1 import (
    OverviewListingPaginator,
    CommentsListingPaginator,
    SubmittedListingPaginator,
    GildedListingPaginator,
    UpvotedListingPaginator,
    DownvotedListingPaginator,
    HiddenListingPaginator,
    SavedListingPaginator,
    SavedSubmissionsListingPaginator,
    SavedCommentsListingPaginator,
)

class Pull:
    def __init__(self, client: Client) -> None:
        self._client = client

    def overview(self, name: str, amount: Optional[int] = None, *,
            sort: str = 'new',
            ) -> ImpartedPaginatorChainingIterator[OverviewListingPaginator, object]:
        p = OverviewListingPaginator(self._client, f'/user/{name}/overview')
        p.sort = sort
        return ImpartedPaginatorChainingIterator(p, amount)

    def comments(self, name: str, amount: Optional[int] = None, *,
            sort: str = 'new',
            ) -> ImpartedPaginatorChainingIterator[CommentsListingPaginator, LooseComment]:
        p = CommentsListingPaginator(self._client, f'/user/{name}/comments')
        p.sort = sort
        return ImpartedPaginatorChainingIterator(p, amount)

    def submitted(self, name: str, amount: Optional[int] = None, *,
            sort: str = 'hot',
            ) -> ImpartedPaginatorChainingIterator[SubmittedListingPaginator, Submission]:
        p = SubmittedListingPaginator(self._client, f'/user/{name}/submitted')
        p.sort = sort
        return ImpartedPaginatorChainingIterator(p, amount)

    def awards_received(self, name: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[GildedListingPaginator, object]:
        p = GildedListingPaginator(self._client, f'/user/{name}/gilded')
        return ImpartedPaginatorChainingIterator(p, amount)

    def awards_given(self, name: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[GildedListingPaginator, object]:
        p = GildedListingPaginator(self._client, f'/user/{name}/gilded/given')
        return ImpartedPaginatorChainingIterator(p, amount)

    def upvoted(self, name: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[UpvotedListingPaginator, Submission]:
        p = UpvotedListingPaginator(self._client, f'/user/{name}/upvoted')
        return ImpartedPaginatorChainingIterator(p, amount)

    def downvoted(self, name: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[DownvotedListingPaginator, Submission]:
        p = DownvotedListingPaginator(self._client, f'/user/{name}/downvoted')
        return ImpartedPaginatorChainingIterator(p, amount)

    def hidden(self, name: str, amount: Optional[int] = None,
            ) -> ImpartedPaginatorChainingIterator[HiddenListingPaginator, Submission]:
        p = HiddenListingPaginator(self._client, f'/user/{name}/hidden')
        return ImpartedPaginatorChainingIterator(p, amount)

    class Saved:
        def __init__(self, outer: Pull) -> None:
            self._client = outer._client

        def __call__(self, name: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[SavedListingPaginator, object]:
            p = SavedListingPaginator(self._client, f'/user/{name}/saved')
            return ImpartedPaginatorChainingIterator(p, amount)

        def submissions(self, name: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[SavedSubmissionsListingPaginator, Submission]:
            p = SavedSubmissionsListingPaginator(self._client, f'/user/{name}/saved')
            return ImpartedPaginatorChainingIterator(p, amount)

        def comments(self, name: str, amount: Optional[int] = None,
                ) -> ImpartedPaginatorChainingIterator[SavedCommentsListingPaginator, LooseComment]:
            p = SavedCommentsListingPaginator(self._client, f'/user/{name}/saved')
            return ImpartedPaginatorChainingIterator(p, amount)

    saved: cached_property[Saved] = cached_property(Saved)




from __future__ import annotations
from typing import TYPE_CHECKING, Any, Callable
if TYPE_CHECKING:
    from ..response import Response

import json

json_decoder: json.JSONDecoder = json.JSONDecoder()
json_decode: Callable[[str], Any] = json_decoder.decode

def load_json_from_response(resp: Response, /) -> Any:
    """Return the JSON data contained in a response object.

    .. .PARAMETERS

    :param response:
    :type response: :class:`~.http.response.Response`

    .. .RETURNS

    :rtype: `Any`

    .. .RAISES

    :raises ValueError:
        The response does not contain valid JSON data.
    """
    return json_decode(resp.data.decode())

def load_json_from_response_but_prefer_status_code_exception_on_failure(resp: Response, /) -> Any:
    """Return the JSON data contained in a response object.

    Behaves similarly to :func:`.load_json_from_response` but prefers to raise a
    :class:`~redditwarp.http.exceptions.StatusCodeException`
    instead of a `ValueError` when the response body does not contain JSON.
    """
    try:
        return load_json_from_response(resp)
    except ValueError as cause:
        try:
            resp.ensure_successful_status()
        except Exception as exc:
            raise exc from cause
        raise
